# [1] http://gaussian.com/thermo/
# [2] https://doi.org/10.1002/chem.201200497
# [3] https://doi.org/10.1021/acs.organomet.8b00456
# [4] https://cccbdb.nist.gov/thermo.asp
# [5] https://pubs.acs.org/doi/10.1021/acs.jctc.0c01306
#     Single-Point Hessian Calculations
#     Spicher, Grimme, 2021
# [6] https://doi.org/10.1063/5.0061187
#     Calculation of improved enthalpy and entropy of vaporization by a modified
#     partition function in quantum cluster equilibrium theory
#     Ingenmey, Grimme, 2021


from collections import namedtuple
from typing import Literal, Optional, Tuple

import numpy as np
from numpy.typing import ArrayLike, NDArray

from thermoanalysis.constants import (
    C,
    KB,
    KBAU,
    NA,
    R,
    PLANCK,
    J2AU,
    J2CAL,
    AMU2KG,
)
from thermoanalysis.config import p_DEFAULT, ROTOR_CUT_DEFAULT
from thermoanalysis.QCData import QCData


TRANS_ENTROPY_KINDS = Literal["sackur", "sackur_simple"]
VIB_KINDS = Literal["qrrho", "rrho"]


ThermoResults = namedtuple(
    "ThermoResults",
    (
        "T kBT M p org_wavenumbers wavenumbers "
        "scale_factor zpe_scale_factor invert_imag cutoff kind "
        "atom_num linear point_group sym_num "
        "Q_el Q_trans Q_rot Q_vib Q_vib_V0 Q_tot Q_tot_V0 "
        "U_el U_trans U_rot U_vib U_therm U_tot ZPE H "
        "S_trans S_rot S_vib S_el S_tot "
        "c_el c_trans c_rot c_vib c_tot "
        "TS_trans TS_rot TS_vib TS_el TS_tot G dG"
    ),
)


def chai_head_gordon_weights(
    frequencies: NDArray[np.float64], cutoff: float, alpha: int = 4
) -> NDArray[np.float64]:
    """Chai-Head-Gordon damping function.

    Used for interpolating between harmonic oscillator and hindered rotor
    approximations. See eq. (8) in [2], or Eq. (10) in [7].

    Parameters
    ----------
    frequencies
        Vibrational frequencies in 1/s.
    cutoff
        Wavenumber cutoff in cm⁻¹. Vibrations below this threshold will mostly
        be treated as hindered rotors.
    alpha
        Exponent alpha in the damping function.

    Returns
    -------
    weights
        Weights of the damping function.
    """

    wavenumbers = (frequencies / C) / 100  # in cm⁻¹
    weights = 1 / (1 + (cutoff / wavenumbers) ** alpha)
    return weights


##############
# ELECTRONIC #
##############


def electronic_part_func(
    multiplicity: int, electronic_energies: ArrayLike, temperature: float
) -> float:
    """Electronic partition function.

    Parameters
    ----------
    multiplicity
        Multiplicity of the molecule.
    electronic_energies
        Electronic energy/energies in Hartree.
    temperature : float
        Absolute temperature in Kelvin.

    Returns
    -------
    Q_el
        Electronic partition function.
    """

    electronic_energies = np.atleast_1d(electronic_energies)
    energy_diffs = electronic_energies - electronic_energies.min()
    g = multiplicity
    q_el = g * (np.exp(-energy_diffs / (KBAU * temperature))).sum()
    return q_el


def electronic_entropy(multiplicity: int) -> float:
    """Electronic entropy.

    Currently, only one electronic state is considered. See [1] for reference.

    Parameters
    ----------
    multiplicity
        Multiplicity of the molecule.

    Returns
    -------
    S_el
        Electronic entropy in Hartree / particle.
    """
    S_el = KBAU * np.log(multiplicity)
    return S_el


#######################
# TRANSLATIONAL TERMS #
#######################


def translational_part_func(
    molecular_mass: float, temperature: float, pressure: float
) -> float:
    """Translational partition function Q_trans.

    Parameters
    ----------
    molecular_mass
        Molecular mass in atomic mass units (amu).
    temperature
        Absolute temperature in Kelvin.
    pressure
        Pressure in Pascal.

    Returns
    -------
    q_trans
        Translational partition function.
    """
    # Volume of an atom of an ideal gas
    volume = KB * temperature / pressure
    molecular_mass_kg = molecular_mass * AMU2KG
    q_trans = (
        (2 * np.pi * molecular_mass_kg * KB * temperature / PLANCK**2) ** 1.5
    ) * volume
    return q_trans


def translational_energy(temperature: float) -> float:
    """Kinectic energy of an ideal gas.

    See [1] for reference.

    Parameters
    ----------
    temperature
        Absolute temperature in Kelvin.

    Returns
    -------
    U_trans
        Kinetic energy in Hartree / particle.
    """
    U_trans = 3 / 2 * KBAU * temperature
    return U_trans


def sackur_tetrode(molecular_mass: float, temperature: float, pressure: float) -> float:
    """Translational entropy of an ideal gas.

    See [1] for reference.

    Parameters
    ----------
    molecular_mass
        Molecular mass in atomic mass units (amu).
    temperature
        Absolute temperature in Kelvin.
    pressure
        Pressure in Pascal.

    Returns
    -------
    S_trans
        Translational entropy in Hartree / (particle * K).
    """
    # Just using 1e5 instead of a "true" atmosphere of 1.01325e5 seems to
    # agree better with the results Gaussian and ORCA produce.
    q_trans = (
        (2 * np.pi * molecular_mass * AMU2KG * KB * temperature / PLANCK**2) ** (3 / 2)
        * KB
        * temperature
        / pressure
    )
    S_trans = KBAU * (np.log(q_trans) + 1 + 3 / 2)
    return S_trans


def sackur_tetrode_simplified(molecular_mass: float, temperature: float) -> float:
    """Translational entropy of a monoatomic ideal gas.

    See [3] for reference.

    Parameters
    ----------
    molecular_mass
        Molecular mass in atomic mass units (amu).
    temperature
        Absolute temperature in Kelvin.

    Returns
    -------
    S_trans
        Translational entropy in J/(mol*K).
    """
    S_trans = (
        3 / 2 * R * np.log(molecular_mass) + 5 / 2 * R * np.log(temperature) - 2.315
    )
    return S_trans


def translational_entropy(
    molecular_mass: float,
    temperature: float,
    pressure: float,
    kind: TRANS_ENTROPY_KINDS = "sackur",
) -> float:
    """Wrapper for translational entropy calculation.

    Parameters
    ----------
    molecular_mass
        Molecular mass in atomic mass units (amu).
    temperature
        Absolute temperature in Kelvin.
    kind
        Type of calculation method.

    Returns
    -------
    S_trans
        Translational entropy.
    """
    funcs = {
        "sackur": lambda M, T: sackur_tetrode(
            molecular_mass=M, temperature=T, pressure=pressure
        ),
        "sackur_simple": sackur_tetrode_simplified,
    }
    return funcs[kind](molecular_mass, temperature)


def translational_heat_capacity() -> float:
    """Constant volume heat capacity.

    Returns
    -------
    c_trans
        Constant volume heat capacity in J / (K mol).
    """
    c_trans = 1.5 * R
    return c_trans


####################
# ROTATIONAL TERMS #
####################


def rotational_part_func(
    temperature: float,
    rot_temperatures: NDArray[np.float64],
    symmetry_number: int,
    is_atom: bool,
    is_linear: bool,
) -> float:
    """Rotational partition function Q_rot.

    See [1] for reference.

    Parameters
    ----------
    temperature
        Absolute temperature in Kelvin.
    rot_temperatures
        Rotational temperatures in Kelvin.
    symmetry_number
        Symmetry number.
    is_atom
        Wether the molcule is an atom.
    is_linear
        Wether the molecule is linear.

    Returns
    -------
    Q_rot
        Rotational partition function.
    """

    if is_atom:
        q_rot = 1
    elif is_linear:
        # First rot_temperature will be infinite, and the last two components
        # will be identical. Only use the last one.
        q_rot = temperature / (rot_temperatures[-1] * symmetry_number)
    else:
        # Polyamtomic, non-linear case
        q_rot = (
            np.pi ** (1 / 2)
            / symmetry_number
            * (temperature ** (3 / 2) / np.prod(rot_temperatures) ** (1 / 2))
        )
    return q_rot


def rotational_energy(temperature: float, is_linear: bool, is_atom: bool) -> float:
    """Rotational energy.

    See [1] for reference.

    Parameters
    ----------
    temperature
        Absolute temperature in Kelvin.
    is_linear
        Wether the molecule is linear.
    is_atom
        Wether the molcule is an atom.

    Returns
    -------
    U_rot
        Rotational energy in Hartree / particle.
    """
    if is_atom:
        factor = 0.0
    elif is_linear:
        factor = 1.0
    else:
        factor = 1.5

    U_rot = factor * KBAU * temperature
    return U_rot


def rotational_entropy(
    temperature: float,
    rot_temperatures: NDArray[np.float64],
    symmetry_number: int,
    is_atom: bool,
    is_linear: bool,
) -> float:
    """Rotational entropy.

    See [1] for reference.

    Parameters
    ----------
    temperature
        Absolute temperature in Kelvin.
    rot_temperatures
        Rotational temperatures in Kelvin.
    symmetry_number
        Symmetry number.
    is_atom
        Wether the molcule is an atom.
    is_linear
        Wether the molecule is linear.

    Returns
    -------
    S_rot
        Rotational entropy in Hartree /(particle * K).
    """

    q_rot = rotational_part_func(
        temperature, rot_temperatures, symmetry_number, is_atom, is_linear=is_linear
    )
    if is_atom:
        plus = 0.0
    elif is_linear:
        plus = 1.0
    # Polyamtomic, non-linear case
    else:
        plus = 1.5
    S_rot = KBAU * (np.log(q_rot) + plus)
    return S_rot


def rotational_heat_capacity(is_atom: bool, is_linear: bool) -> float:
    """Rotational heat capacity.

    Parameters
    ----------
    is_atom
        Wether the molcule is an atom.
    is_linear
        Wether the molecule is linear.

    Returns
    -------
    c_rot
        Rotational contributions to the heat capacity.
    """
    if is_atom:
        c_rot = 0
    elif is_linear:
        c_rot = R
    else:
        c_rot = 1.5 * R
    return c_rot


#####################
# VIBRATIONAL TERMS #
#####################


def zero_point_energy(frequencies: NDArray[np.float64]) -> float:
    """Vibrational zero point energies.

    See [1] for reference.

    Parameters
    ----------
    frequencies
        Vibrational frequencies in 1/s.

    Returns
    -------
    ZPE
        Vibrational ZPE for the given frequencies in Hartree / particle.
    """

    ZPE = J2AU * (PLANCK * frequencies / 2).sum()
    return ZPE


def vibrational_part_funcs(
    temperature: float, frequencies: NDArray[np.float64]
) -> Tuple[NDArray[np.float64], NDArray[np.float64]]:
    """Vibrational partition functions for harmonic oscillators.

    Parameters
    ----------
    temperature
        Absolute temperature in Kelvin.
    frequencies
        Vibrational frequencies in 1/s.

    Returns
    -------
    q_vibs : ArrayLike
        Vibrational partition functions at the bottom of the well.
    q_vibs_V0  : ArrayLike
        Vibrational partition functions with first vibrational energy level
        as reference.
    """
    frequencies = np.array(frequencies, dtype=float)
    quot = -PLANCK * frequencies / (KB * temperature)
    denom = 1 / (1 - np.exp(quot))
    quot_half = quot / 2
    # Bottom of the well as reference
    q_vibs = np.exp(quot_half) * denom
    # First vibrational energy level as reference
    q_vibs_V0 = denom
    return q_vibs, q_vibs_V0


def qrrho_vibrational_part_func(
    temperature: float,
    frequencies: NDArray[np.float64],
    I_mean: Optional[float],
    cutoff: float,
    alpha: int = 4,
) -> Tuple[float, float]:
    """QRRHO vibrational partition function.

    As given in Eq. (7) in [6]. Mix between hindererd rotor
    and harmonic oscillator partition function.

    Parameters
    ----------
    temperature
        Absolute temperature in Kelvin.
    frequencies
        Vibrational frequencies in 1/s.
    I_mean
        Average moment of inertia of the molecule in Angstrom² * amu.
    cutoff
        Wavenumber cutoff in cm⁻¹. Vibrations below this threshold will mostly
        be treated as hindered rotors.
    alpha
        Exponent alpha in the damping function (Eq. (8) in [2], or Eq. (10) in [7])

    Returns
    -------
    q_qrrho
        QRRHO Vibrational partition function, unitless.
    q_qrrho_V0
        QRRHO Vibrational partition function, evaluated at first vibrational energy level.
    """

    wavenumbers = (frequencies / C) / 100  # in cm⁻¹
    q_vibs, q_vibs_V0 = vibrational_part_funcs(temperature, frequencies)

    if cutoff > 0.0 and I_mean:
        # Normal mode moments of inertia
        wavenumbers_m = 100 * wavenumbers  # in m⁻¹
        mu = PLANCK / (
            2 * np.pi * 4 * np.pi * wavenumbers_m
        )  # Eq. (9) in [6], in kg m³ s⁻¹
        # Convert average moment of inertia from Å² AMU to SI units (m² kg)
        I_mean_SI = I_mean * 1e-20 * AMU2KG
        mu_ = mu * I_mean_SI / (mu + I_mean_SI)
        q_hr = np.sqrt(
            2 * mu_ / (np.pi * (PLANCK / (2 * np.pi)) ** 2 / (temperature * KB))
        )
    # Without cutoff this partition function reduces to the partition function
    # of the harmonic oscillator.
    else:
        q_hr = np.ones_like(q_vibs)

    weights = chai_head_gordon_weights(frequencies, cutoff, alpha)

    def prod(q_vibs):
        # Work in log-domain to avoid overflow in large mode products.
        tiny = np.finfo(float).tiny
        max_log = np.log(np.finfo(float).max)
        min_log = np.log(tiny)
        log_q = np.sum(
            weights * np.log(np.clip(q_vibs, tiny, None))
            + (1 - weights) * np.log(np.clip(q_hr, tiny, None))
        )
        if log_q >= max_log:
            return float("inf")
        if log_q <= min_log:
            return 0.0
        return float(np.exp(log_q))

    q_qrrho = prod(q_vibs)
    q_qrrho_V0 = prod(q_vibs_V0)
    return q_qrrho, q_qrrho_V0


def vibrational_part_func(
    temperature: float, frequencies: NDArray[np.float64]
) -> Tuple[float, float]:
    """Vibrational partition function.

    Parameters
    ----------
    temperature
        Absolute temperature in Kelvin.
    frequencies
        Vibrational frequencies in 1/s.

    Returns
    -------
    q_qrrho
        QRRHO Vibrational partition function, unitless.
    q_qrrho_V0
        QRRHO Vibrational partition function, evaluated at first vibrational energy level,
        unitless.
    """
    return qrrho_vibrational_part_func(
        temperature, frequencies, I_mean=None, cutoff=0.0
    )


def vibrational_energy(temperature: float, frequencies: NDArray[np.float64]) -> float:
    """Vibrational energy.

    See [1] for reference.

    Parameters
    ----------
    temperature
        Absolute temperature in Kelvin.
    frequencies
        Vibrational frequencies in 1/s.

    Returns
    -------
    U_vib
        Vibrational energy in Hartree / particle.
    """
    vib_temperatures = PLANCK * frequencies / KB
    U_vib = KBAU * np.sum(
        vib_temperatures * (1 / 2 + 1 / (np.exp(vib_temperatures / temperature) - 1))
    )
    return U_vib


def harmonic_vibrational_entropies(
    temperature: float, frequencies: NDArray[np.float64]
) -> NDArray[np.float64]:
    """Vibrational entropy of a harmonic oscillator.

    See [1] and [2] for reference. Eq. (3) in the Grimme paper
    is lacking a T in the denominator of the first term. It is given as
    h*w/(k(e^(hw/kt) -1)) but it must be h*w(kT(e^(hw/kT)-1)) instead.
    Here the calculation is done as presented in [1].

    Parameters
    ----------
    temperature : float
        Absolute temperature in Kelvin.
    frequencies : array-like
        Vibrational frequencies in 1/s.

    Returns
    -------
    S_vib : array-like
        Array containing vibrational entropies in Hartree / (particle * K).
    """

    # Correct formula from the Grimme paper [3].
    # hnu = PLANCK * frequencies
    # hnu_kt = hnu / (KB * temperature)
    # S_vib = KB * (hnu / (KB*(np.exp(hnu_kt) - 1)*temperature)
    # - np.log(1 - np.exp(-hnu_kt))
    # ).sum()

    # As given in [1].
    vib_temps = frequencies * PLANCK / KB
    S_vibs = KBAU * (
        (vib_temps / temperature) / (np.exp(vib_temps / temperature) - 1)
        - np.log(1 - np.exp(-vib_temps / temperature))
    )
    return S_vibs


def free_rotor_entropies(
    temperature: float, frequencies: NDArray[np.float64], B_av: float = 1e-44
) -> NDArray[np.float64]:
    """Entropy of a free rotor.

    See [2] for reference.

    Parameters
    ----------
    temperature
        Absolute temperature in Kelvin.
    frequencies
        Vibrational frequencies in 1/s.
    B_av
        Limiting value for effective moment of inertia in kg*m².

    Returns
    -------
    S_free_rots
        Array containing free-rotor entropies in Hartree / (particle * K).
    """
    inertia_moments = PLANCK / (8 * np.pi**2 * frequencies)
    eff_inertia_moments = (inertia_moments * B_av) / (inertia_moments + B_av)
    S_free_rots = KBAU * (
        1 / 2
        + np.log(
            (8 * np.pi**3 * eff_inertia_moments * KB * temperature / PLANCK**2)
            ** (1 / 2)
        )
    )
    return S_free_rots


def vibrational_entropies(
    temperature: float, frequencies: NDArray[np.float64], cutoff: float, alpha: int = 4
) -> NDArray[np.float64]:
    """Weighted vibrational entropy.

    As given in Eq. (7) of [2].

    Parameters
    ----------
    temperature
        Absolute temperature in Kelvin.
    frequencies
        Vibrational frequencies in 1/s.
    cutoff
        Wavenumber cutoff in cm⁻¹. Vibrations below this threshold will mostly
        be described by free-rotor entropies.
    alpha
        Exponent alpha in the damping function (Eq. (8) in [2]).

    Returns
    -------
    S_vibs
        Array containing vibrational entropies in Hartree / (particle * K).
    """
    weights = chai_head_gordon_weights(frequencies, cutoff, alpha)
    S_harmonic = harmonic_vibrational_entropies(temperature, frequencies)
    S_quasi_harmonic = free_rotor_entropies(temperature, frequencies)
    S_vibs = weights * S_harmonic + (1 - weights) * S_quasi_harmonic
    return S_vibs


def vibrational_entropy(
    temperature: float, frequencies: NDArray[np.float64], cutoff: float, alpha: int = 4
) -> float:
    """Vibrational entropy.

    Wrapper function. As given in Eq. (7) of [2].

    Parameters
    ----------
    temperature
        Absolute temperature in Kelvin.
    frequencies
        Vibrational frequencies in 1/s.
    cutoff
        Wavenumber cutoff in cm⁻¹. Vibrations below this threshold will mostly
        be described by free-rotor entropies.
    alpha
        Exponent alpha in the damping function (Eq. (8) in [2]).

    Returns
    -------
    S_vib
        Vibrational entropy in Hartree / (particle * K).
    """
    return vibrational_entropies(temperature, frequencies, cutoff, alpha).sum()


def vibrational_heat_capacity(
    temperature: float, frequencies: NDArray[np.float64]
) -> float:
    """
    Parameters
    ----------
    temperature
        Absolute temperature in Kelvin.
    frequencies
        Vibrational frequencies in 1/s.

    Returns
    -------
    c_vib
        Vibrational contributions to the heat capacity in J / (K mol).
    """
    quot = PLANCK * frequencies / (KB * temperature)

    c_vib = R * (np.exp(quot) * (quot / (np.exp(quot) - 1)) ** 2).sum()
    return c_vib


def thermochemistry(
    qc: QCData,
    temperature: float,
    pressure: float = p_DEFAULT,
    kind: VIB_KINDS = "qrrho",
    rotor_cutoff: float = ROTOR_CUT_DEFAULT,
    scale_factor: float = 1.0,
    zpe_scale_factor: float = 1.0,
    invert_imags: float = 0.0,
    cutoff: float = 0.0,
) -> ThermoResults:
    assert kind in "qrrho rrho".split()
    assert invert_imags <= 0.0
    assert cutoff >= 0.0

    T = temperature
    pressure = pressure

    org_wavenumbers = qc.wavenumbers.copy()
    wavenumbers = org_wavenumbers.copy()
    # If given a full set of 3N normal mode wavenumbers, exclude N wavenumbers
    # with the smallest absolute wavenumbers.
    if wavenumbers.size == 3 * qc.atom_num:
        abs_wavenumbers = np.abs(wavenumbers)
        inds = np.argsort(abs_wavenumbers)
        drop_first = 6 - int(qc.is_linear)
        # Drop first N wavenumbers with smallest absolute values
        keep_mask = np.ones_like(wavenumbers, dtype=bool)
        keep_mask[inds[:drop_first]] = False
        wavenumbers = wavenumbers[keep_mask]

    """
    Remaining imaginary frequencies with small absolute values can be inverted.
    Afterwards, all frequencies are scaled by the given factor. Finally, positive
    frequencies below a given cutoff are set to this cutoff, e.g. 50 cm⁻¹.

    [5] suggests inverting imaginary frequencies above -20 cm⁻¹.
    """
    if invert_imags < 0.0:
        to_invert = np.logical_and(invert_imags <= wavenumbers, wavenumbers <= 0.0)
        if to_invert.sum() > 0:
            print(f"Inverted {to_invert.sum()} imaginary frequencies.")
        wavenumbers[to_invert] *= -1

    wavenumbers *= scale_factor

    if cutoff > 0.0:
        to_cut = np.logical_and(wavenumbers < cutoff, wavenumbers > 0.0)
        wavenumbers[to_cut] = cutoff

    wavenumbers = wavenumbers[wavenumbers > 0.0]
    vib_frequencies = C * wavenumbers * 100

    # Partition functions
    Q_el = electronic_part_func(qc.mult, qc.scf_energy, temperature=T)
    Q_trans = translational_part_func(qc.M, temperature=T, pressure=pressure)
    Q_rot = rotational_part_func(
        temperature=T,
        rot_temperatures=qc.rot_temperatures,
        symmetry_number=qc.symmetry_number,
        is_atom=qc.is_atom,
        is_linear=qc.is_linear,
    )

    # Zero point energy
    zpe_org = zpe_scale_factor * zero_point_energy(vib_frequencies)
    zpe = zpe_scale_factor * zpe_org

    # Internal energies
    U_el = qc.scf_energy
    U_trans = translational_energy(T)
    U_rot = rotational_energy(T, qc.is_linear, qc.is_atom)
    # U_vib already includes ZPE
    U_vib = vibrational_energy(T, vib_frequencies)
    # Replace ZPE in U_vib w/ scaled ZPE
    U_vib = U_vib - zpe_org + zpe
    U_therm = U_rot + U_vib + U_trans
    U_tot = U_el + U_therm

    H = U_tot + KBAU * T

    # Entropies
    S_el = electronic_entropy(qc.mult)
    S_rot = rotational_entropy(
        T,
        qc.rot_temperatures,
        qc.symmetry_number,
        is_atom=qc.is_atom,
        is_linear=qc.is_linear,
    )
    S_trans = translational_entropy(qc.M, T, pressure=pressure)

    # Heat capacities
    c_el = 0.0  # always zero
    c_trans = translational_heat_capacity()
    c_rot = rotational_heat_capacity(is_atom=qc.is_atom, is_linear=qc.is_linear)
    c_vib = vibrational_heat_capacity(T, vib_frequencies)
    c_tot = c_el + c_trans + c_rot + c_vib

    if kind == "rrho":
        S_hvibs = harmonic_vibrational_entropies(T, vib_frequencies)
        S_vib = S_hvibs.sum()
        # Harmonic oscillator partition function
        Q_vib, Q_vib_V0 = vibrational_part_func(
            frequencies=vib_frequencies, temperature=T
        )
    elif kind == "qrrho":
        S_vib = vibrational_entropy(T, vib_frequencies, cutoff=rotor_cutoff)
        I_mean = qc.average_moment_of_inertia
        Q_vib, Q_vib_V0 = qrrho_vibrational_part_func(
            T, vib_frequencies, I_mean, cutoff=rotor_cutoff
        )
    else:
        raise Exception("You should never get here!")

    Q_tot = Q_el * Q_trans * Q_rot * Q_vib
    Q_tot_V0 = Q_el * Q_trans * Q_rot * Q_vib_V0

    S_tot = S_el + S_trans + S_rot + S_vib
    G = H - T * S_tot
    dG = G - U_el

    thermo = ThermoResults(
        T=temperature,
        kBT=KBAU * temperature,
        M=qc.M,
        p=pressure,
        point_group=qc.point_group,
        sym_num=qc.symmetry_number,
        scale_factor=scale_factor,
        zpe_scale_factor=zpe_scale_factor,
        invert_imag=invert_imags,
        cutoff=cutoff,
        kind=kind,
        org_wavenumbers=org_wavenumbers,
        wavenumbers=wavenumbers,
        atom_num=qc.atom_num,
        linear=qc.is_linear,
        Q_el=Q_el,
        Q_trans=Q_trans,
        Q_rot=Q_rot,
        Q_vib=Q_vib,
        Q_vib_V0=Q_vib_V0,
        Q_tot=Q_tot,
        Q_tot_V0=Q_tot_V0,
        U_el=U_el,
        U_trans=U_trans,
        U_rot=U_rot,
        U_vib=U_vib,
        U_therm=U_therm,
        U_tot=U_tot,
        ZPE=zpe,
        H=H,
        S_trans=S_trans,
        S_rot=S_rot,
        S_vib=S_vib,
        S_el=S_el,
        S_tot=S_tot,
        c_el=c_el,
        c_trans=c_trans,
        c_rot=c_rot,
        c_vib=c_vib,
        c_tot=c_tot,
        TS_trans=T * S_trans,
        TS_rot=T * S_rot,
        TS_vib=T * S_vib,
        TS_el=T * S_el,
        TS_tot=T * S_tot,
        G=G,
        dG=dG,
    )
    return thermo


def print_thermo_results(thermo_results: ThermoResults):
    au2CalMol = 1 / J2AU * NA * J2CAL

    def toCalMol(E):
        return f"{E*au2CalMol: >14.2f} cal/mol"

    def StoCalKMol(S):
        return f"{S*au2CalMol: >12.2f} cal/(K mol)"

    def CtoCalKMol(c):
        return f"{c*J2CAL: >12.4f} cal/(K mol)"

    def part_func(Q):
        return f"{Q: >12.8e}"

    def print_line(key, fmt_func, num):
        print(f"{key: >18s} = {fmt_func(num)}")

    def pQ(key, num):
        print_line(f"Q_{key}", part_func, num)

    def pU(key, num):
        print_line(f"U_{key}", toCalMol, num)

    def pS(key, num):
        print_line(f"S_{key}", StoCalKMol, num)

    def pC(key, num):
        print_line(f"C_{key}", CtoCalKMol, num)

    tr = thermo_results
    T = tr.T
    print(
        f"Thermochemistry @ {T:.2f} K and {tr.p:.6e} Pa, '{tr.kind.upper()}' analysis.\n"
    )

    print("Partition functions:")
    pQ("el", tr.Q_el)
    pQ("trans", tr.Q_trans)
    pQ("rot", tr.Q_rot)
    pQ("vib BOT", tr.Q_vib)
    pQ("vib V0", tr.Q_vib_V0)
    print()

    print("Zero-point energy")
    print_line("ZPE", toCalMol, tr.ZPE)
    print()

    print("Internal energy")
    pU("el", tr.U_el)
    pU("trans", tr.U_trans)
    pU("rot", tr.U_rot)
    pU("vib (incl. ZPE)", tr.U_vib)
    pU("therm", tr.U_therm)
    print("U_tot = U_el + U_trans + U_rot + U_vib")
    pU("tot", tr.U_tot)
    print()

    print("Entropy contributions:")
    pS("el", tr.S_el)
    pS("trans", tr.S_trans)
    pS("rot", tr.S_rot)
    pS("vib", tr.S_vib)
    print("S_tot = S_el + S_trans + S_rot + S_vib")
    pS("tot", tr.S_tot)
    print()

    print("Heat capacities:")
    pC("el", tr.c_el)
    pC("trans", tr.c_trans)
    pC("rot", tr.c_rot)
    pC("vib", tr.c_vib)
    pC("tot", tr.c_tot)



# [1] https://pubs.acs.org/doi/pdf/10.1021/ct200290m?rand=dcfwsf09
# [2] https://onlinelibrary.wiley.com/doi/epdf/10.1002/jcc.23481
# [3] https://onlinelibrary.wiley.com/doi/epdf/10.1002/tcr.201600043

import itertools as it
from math import isclose
from typing import List

import autograd
import autograd.numpy as anp
import numpy as np
import numpy.typing as npt

from pysisyphus.calculators.Calculator import Calculator
from pysisyphus.elem_data import COVALENT_RADII
from pysisyphus.Geometry import Geometry
from pysisyphus.helpers import complete_fragments
from pysisyphus.helpers_pure import log
from pysisyphus.io.hdf5 import get_h5_group, resize_h5_group
from pysisyphus.linalg import finite_difference_hessian


def get_data_model(atoms, max_cycles):
    coord_size = 3 * len(atoms)
    _1d = (max_cycles,)
    _2d = (max_cycles, coord_size)
    _3d = (max_cycles, coord_size, coord_size)

    data_model = {
        "cart_coords": _2d,
        "energy": _1d,
        "forces": _2d,
        "hessian": _3d,
        "true_energy": _1d,
        "true_forces": _2d,
        "true_hessian": _3d,
    }

    return data_model


class CovRadiiSumZero(Exception):
    pass


def afir_closure(
    fragment_indices, cov_radii, gamma, rho=1, p=6, prefactor=1.0, logger=None
):
    """rho=1 pushes fragments together, rho=-1 pulls fragments apart."""

    # See https://onlinelibrary.wiley.com/doi/full/10.1002/qua.24757
    # Eq. (9) for extension to 3 fragments.
    assert len(fragment_indices) == 2

    inds = np.array(list(it.product(*fragment_indices)))
    cov_rad_sums = cov_radii[inds].sum(axis=1)

    if isclose(cov_rad_sums.sum(), 0.0, abs_tol=1e-10):
        raise CovRadiiSumZero("Sum of covalent radii is 0.0!")

    # 3.8164 Angstrom in Bohr
    R0 = 7.21195
    # 1.0061 kJ/mol to Hartree
    epsilon = 0.000383203368

    # Avoid division by zero for gamma = 0.
    if gamma == 0.0:
        alpha = 0.0
    else:
        alpha = gamma / (
            (2 ** (-1 / 6) - (1 + (1 + gamma / epsilon) ** 0.5) ** (-1 / 6)) * R0
        )

    rho_verbose = {1: ("pushing", "together"), -1: ("pulling", "apart")}
    w1, w2 = rho_verbose[rho]
    log(
        logger,
        f"Creating AFIR closure with α={alpha:.6f}, prefactor {prefactor:.6f}, "
        f"rho={rho}, {w1} framgents {w2}",
    )

    def afir_func(coords3d):
        diffs = anp.diff(coords3d[inds], axis=1).reshape(-1, 3)
        rs = anp.linalg.norm(diffs, axis=1)

        omegas = (cov_rad_sums / rs) ** p

        f = prefactor * alpha * rho * (omegas * rs).sum() / omegas.sum()
        return f

    return afir_func


class AFIR(Calculator):
    def __init__(
        self,
        calculator: Calculator,
        fragment_indices: List[List[int]],
        gamma: npt.ArrayLike,
        rho: npt.ArrayLike = 1,
        p: int = 6,
        ignore_hydrogen: bool = False,
        zero_hydrogen: bool = True,
        complete_fragments: bool = True,
        dump: bool = True,
        h5_fn: str = "afir.h5",
        h5_group_name: str = "afir",
        **kwargs,
    ):
        """
        Artifical Force Induced Reaction calculator.

        Currently, there are no automated drivers to run large-scale AFIR calculations
        with many different initial orientations and/or increasing collision energy
        parameter γ. Nontheless, selected AFIR calculations can be carried out by hand.
        After convergence, artificial potential & forces, as well as real energies
        and forces can be plotted with 'pysisplot --afir'. The highest energy point
        along the AFIR path can then be selected for a subsequent TS-optimization,
        e.g. via 'pysistrj --get [index] optimzation_trj.xyz'.

        Future versions of pysisyphus may provide drivers for more automatted
        AFIR calculations.

        Parameters
        ----------
        calculator
            Actual QC calculator that provides energies and its derivatives,
            that are modified by the AFIR calculator, e.g., ORCA or Psi4.
        fragment_indices
            List of lists of integers, specifying the separate fragments. If
            the indices in theses lists don't comprise all atoms in the molecule,
            the reamining indices will be added as a separate fragment. If a AFIR
            calculation is carried out with 2 fragments and 'complete_fragments' is
            True (see below) it is enough to specify only
            the indices of one fragment, e.g., for a system of 10 atoms
            'fragment_indices=[[0,1,2,3]]' is enough. The second system will be set
            up automatically with indices [4,5,6,7,8,9].
        gamma
            Collision energy parameter γ in au. For 2 fragments it can be a single
            integer, while for > 2 fragments a list of gammas must be given, specifying
            the pair-wise collision energy parameters. For 3 fragments 3 gammas must be
            given [γ_01, γ_02, γ_12], for 4 fragments 6 gammas would be required
            [γ_01, γ_02, γ_03, γ_12, γ_13, γ_23] and so on.
        rho
            Direction of the artificial force, either 1 or -1. The same comments
            as for gamma apply. For 2 fragments a single integer is enough, for
            > 2 fragments a list of rhos must be given (see above).  For rho=1
            fragments are pushed together, for rho=-1 fragments are pulled apart.
        p
            Exponent p used in the calculation of the weight function ω. Defaults
            to 6 and probably does not have to be changed.
        ignore_hydrogen
            Whether hydrogens are ignored in the calculation of the artificial force.
            All weights between atom pairs containing hydrogen will be set to 0.
        zero_hydrogen
            Whether to use 0.0 as covalent radius for  hydrogen in the weight function.
            Compared to 'ignore_hydrogen', which results in zero weights for all atom
            pairs involving hydrogen, 'zero_hydrogen' may be non-zero, depending on the
            covalent radius of the second atom in the pair.
        complete_fragments
            Whether an incomplete specification in 'fragment_indices' is automatically
            completed.
        dump
            Whether an HDF5 file is created.
        h5_fn
            Filename of the HDF5 file used for dumping.
        h5_group_name
            HDF5 group name used for dumping.

        Other Parameters
        ----------------
        **kwargs
            Keyword arguments passed to the Calculator baseclass.
        """
        super().__init__(**kwargs)

        self.calculator = calculator
        # Update own charge and multiplicity with values from the wrapped
        # calculator.
        self.charge = calculator.charge
        self.mult = calculator.mult
        self.fragment_indices = fragment_indices
        assert len(self.fragment_indices) > 0
        self.gamma = np.atleast_1d(gamma).astype(float)
        self.rho = np.atleast_1d(rho).astype(int)
        np.testing.assert_allclose(np.abs(self.rho), np.ones_like(self.rho))
        self.p = p
        self.ignore_hydrogen = ignore_hydrogen
        self.zero_hydrogen = zero_hydrogen
        self.complete_fragments = complete_fragments
        self.dump = dump
        self.h5_fn = self.out_dir / h5_fn
        self.h5_group_name = h5_group_name
        # We can't initialize the HDF5 group as we don't know the shape of
        # atoms/coords yet. So we wait until after the first calculation.
        self.h5_cycles = 50

        if self.ignore_hydrogen:
            self.log("No artificial force contribution from hydrogens!")
        self.atoms = None
        self.calc_counter = 0
        self.h5_group_initialized = False

    """We try to keep charge and multiplicity consistent between the AFIR
    calculator and the actual wrapped calculator. But we will always return
    the charge and multiplicity of the underlying calculator."""

    @property
    def charge(self):
        return self.calculator.charge

    @charge.setter
    def charge(self, charge):
        try:
            self.calculator.charge = charge
        except AttributeError:
            pass
        self._charge = charge

    @property
    def mult(self):
        return self.calculator.mult

    @mult.setter
    def mult(self, mult):
        try:
            self.calculator.mult = mult
        except AttributeError:
            pass
        self._mult = mult

    def init_h5_group(self, atoms, max_cycles=None):
        if max_cycles is None:
            max_cycles = self.h5_cycles
        self.data_model = get_data_model(atoms, max_cycles)
        # h5_group = get_h5_group(
        #     self.h5_fn, self.h5_group_name, self.data_model, reset=True
        # )
        self.h5_group_initialized = True
        # return h5_group

    def dump_h5(self, atoms, coords, results):
        # Initialize if not done yet
        # if not self.h5_group_initialized:
        #     h5_group = self.init_h5_group(atoms)
        #     # Write atoms only once
        #     h5_group.attrs["atoms"] = atoms

        # h5_group = get_h5_group(self.h5_fn, self.h5_group_name)

        # # Check if HDF5 datasets have to be resized
        # cur_max_cycles = h5_group["cart_coords"].shape[0]
        # need_resize = self.calc_counter > cur_max_cycles - 5
        # if need_resize:
        #     new_max_cycles = cur_max_cycles + self.h5_cycles
        #     resize_h5_group(h5_group, max_cycles=new_max_cycles)

        # for k, v in results.items():
        #     h5_group[k][self.calc_counter] = v
        # h5_group["cart_coords"][self.calc_counter] = coords
        # h5_group.attrs["cur_cycle"] = self.calc_counter
        # h5_group.file.close()
        pass

    def log_fragments(self):
        self.log(f"Using {len(self.fragment_indices)} fragments")
        for i, frag in enumerate(self.fragment_indices):
            self.log(f"Fragment {i:02d}, {len(frag)} atoms:")
            self.log(f"\t{frag}")

    def write_fragment_geoms(self, atoms, coords):
        geom = Geometry(atoms, coords)
        for i, frag in enumerate(self.fragment_indices):
            frag_geom = geom.get_subgeom(frag)
            fn = self.make_fn(f"frag_geom_{i:02d}.xyz")
            with open(fn, "w") as handle:
                handle.write(frag_geom.as_xyz())
            self.log(f"Wrote geometry of fragment {i:02d} to {fn}.")

    def set_atoms_and_funcs(self, atoms, coords):
        """Initially atoms was also an argument to the constructor of AFIR.
        I removed it so creation becomes easier.
        The first time a calculation is requested with a proper atom set
        everything is set up (cov. radii, afir function and corresponding
        gradient). Afterwards there is only a check if atoms != None and it
        is expected that all functions are properly set.

        fragment_indices can also be incomplete w.r.t. to the number of
        atoms. If the sum of the specified fragment atoms is less than the
        number of atoms present then all remaining unspecified atoms will
        be gathered in one fragment.
        """

        if self.atoms is not None:
            assert self.atoms == atoms
            return

        self.log("Setting atoms on AFIR calculator")
        self.atoms = atoms
        if self.complete_fragments:
            self.fragment_indices = complete_fragments(
                self.atoms, self.fragment_indices
            )
        self.log_fragments()
        self.write_fragment_geoms(atoms, coords)

        hydrogen_inds = [i for i, atom in enumerate(atoms) if atom.lower() == "h"]
        if self.ignore_hydrogen:
            fragment_indices = list()
            for i, frag_inds in enumerate(self.fragment_indices):
                frag_inds_no_h = [j for j in frag_inds if j not in hydrogen_inds]
                fragment_indices.append(frag_inds_no_h)
                dropped_hydrogens = len(frag_inds) - len(frag_inds_no_h)
                self.log(f"Ignoring {dropped_hydrogens} hydrogen(s) from fragment {i}.")
            self.fragment_indices = fragment_indices

        self.cov_radii_org = np.array([COVALENT_RADII[atom.lower()] for atom in atoms])
        self.cov_radii = self.cov_radii_org.copy()
        if self.zero_hydrogen:
            self.cov_radii[hydrogen_inds] = 0.0
        self.log("Set covalent radii")
        self.afir_funcs = list()
        self.afir_grad_funcs = list()
        pairs = list(it.combinations(self.fragment_indices, 2))
        prefactor = 1 / len(pairs)
        self.log(
            f"Doing AFIR with {len(pairs)} fragment pairs and prefactor={prefactor:.4f}"
        )
        try:
            self.gamma = [float(self.gamma)] * len(pairs)
        except TypeError:
            assert len(self.gamma) == len(pairs)
        try:
            self.rho = [float(self.rho)] * len(pairs)
        except TypeError:
            assert len(self.rho) == len(pairs)
        self.log(f"Using gamma(s): {self.gamma}")
        self.log(f"  Using rho(s): {self.rho}")

        for (frag1, frag2), gamma, rho in zip(pairs, self.gamma, self.rho):
            afir_func = afir_closure(
                (frag1, frag2),
                self.cov_radii,
                gamma=gamma,
                rho=rho,
                p=self.p,
                prefactor=prefactor,
                logger=self.logger,
            )
            afir_grad_func = autograd.grad(afir_func)
            self.afir_funcs.append(afir_func)
            self.afir_grad_funcs.append(afir_grad_func)
        self.log("Created and set AFIR function & gradient function.")

    def get_energy(self, atoms, coords, **prepare_kwargs):
        self.set_atoms_and_funcs(atoms, coords)

        true_results = self.calculator.get_energy(atoms, coords, **prepare_kwargs)
        true_energy = true_results["energy"]
        coords3d = coords.reshape(-1, 3)
        # Iterate over all fragment pairs
        afir_energy = sum([afir_func(coords3d) for afir_func in self.afir_funcs])
        self.log()

        results = {
            "energy": true_energy + afir_energy,
            "true_energy": true_energy,
        }
        if self.dump:
            self.dump_h5(atoms, coords, results)
        self.calc_counter += 1
        return results

    def get_forces(self, atoms, coords, **prepare_kwargs):
        self.set_atoms_and_funcs(atoms, coords)

        true_results = self.calculator.get_forces(atoms, coords, **prepare_kwargs)
        true_energy = true_results["energy"]
        true_forces = true_results["forces"]

        coords3d = coords.reshape(-1, 3)
        afir_energy = 0.0
        afir_forces = np.zeros_like(coords)
        # Iterate over all fragment pairs
        for afir_func, afir_grad_func in zip(self.afir_funcs, self.afir_grad_funcs):
            afir_energy += afir_func(coords3d)
            # Add negative of the gradient (the force)
            afir_forces += -afir_grad_func(coords3d).flatten()

        true_norm = np.linalg.norm(true_forces)
        afir_norm = np.linalg.norm(afir_forces)
        self.log(
            f"\ntrue_energy={true_energy:.6f} au\n"
            f"afir_energy={afir_energy:.6f} au\n"
            f" sum_energy={true_energy+afir_energy:.6f} au\n"
            f"norm(true_forces)={true_norm:.6f} au/bohr\n"
            f"norm(afir_forces)={afir_norm:.6f} au/bohr\n"
        )

        results = {
            "energy": true_energy + afir_energy,
            "forces": true_forces + afir_forces,
            "true_forces": true_forces,
            "true_energy": true_energy,
        }
        if self.dump:
            self.dump_h5(atoms, coords, results)
        self.calc_counter += 1
        return results

    def afir_fd_hessian_wrapper(self, coords3d, afir_grad_func):
        coords = coords3d.flatten()

        def grad_func(coords):
            afir_grad = afir_grad_func(coords.reshape(-1, 3))
            return afir_grad.flatten()

        fd_hessian = finite_difference_hessian(coords, grad_func, acc=4)
        return fd_hessian

    def get_hessian(self, atoms, coords, **prepare_kwargs):
        self.set_atoms_and_funcs(atoms, coords)

        true_results = self.calculator.get_hessian(atoms, coords, **prepare_kwargs)
        true_energy = true_results["energy"]
        true_hessian = true_results["hessian"]

        coords3d = coords.reshape(-1, 3)
        afir_energy = 0.0
        afir_hessian = np.zeros_like(true_hessian)
        # Iterate over all fragment pairs
        for afir_func, afir_grad_func in zip(self.afir_funcs, self.afir_grad_funcs):
            afir_energy += afir_func(coords3d)
            # AFIR Hessian from finite differences
            afir_hessian += self.afir_fd_hessian_wrapper(coords3d, afir_grad_func)

        results = {
            "energy": true_energy + afir_energy,
            "hessian": true_hessian + afir_hessian,
            "true_hessian": true_hessian,
            "true_energy": true_energy,
        }
        if self.dump:
            self.dump_h5(atoms, coords, results)
        self.calc_counter += 1
        return results

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"AFIR({self.calculator})"



from math import exp, sqrt

import numpy as np
from numpy.random import default_rng

from pysisyphus.constants import KBAU


"""
[1] https://aip.scitation.org/doi/10.1063/1.2408420
[2] https://dx.doi.org/10.1016/j.cpc.2008.01.006
    Reformulation fo the algorithm. This is implemented e.g., in YAFF.
    https://github.com/molmod/yaff/blob/master/yaff/sampling/nvt.py

csvr_closure() is based on the implementation provided on Bussis homepage:
    https://sites.google.com/site/giovannibussi/downloads/resamplekin.tgz
    (At least in my implementation) there seems to be a problem with
    the conserved quantity, which is not conserved at all ...
csvr_closure_2() is based on [2]
"""


RNG = default_rng()


def sum_noises(num, rng=None):
    """
    Parameters
    ----------
    num : int
        Number of independent Gaussian noises to be squared.
    rng : numpy.random.Generator, optional
        Instances of a random number generator (RNG). If it is not provided the module-level
        RNG will be used.
    """

    if rng is None:
        rng = RNG

    if num == 0:
        sum_ = 0.0
    elif num == 1:
        sum_ = rng.normal()**2
    # nn even, dof - 1 odd
    elif (num % 2) == 0:
        sum_ = 2.0 * rng.gamma(shape=num/2)
    # nn odd, dof - 1 even
    else:
        sum_ = 2.0 * rng.gamma(shape=(num-1)/2) + rng.normal()**2

    return sum_


def csvr_closure(sigma, dof, dt, tau=100, rng=None):
    """
    Parameters
    ----------
    sigma : float
        Target average value of the kinetic energy (1/2 dof k_b T) in the same units as
        cur_kinetic_energy.
    dof : int
        Degrees of freedom.
    tau : float
        Timeconstant of the thermostat.    tau : float
        Timeconstant of the thermostat.
    rng : numpy.random.Generator, optional
        Instances of a random number generator (RNG). If it is not provided the module-level
        RNG will be used.
    """
    # Relaxation time of the thermostat in units of "how often this routine
    # is called" (dt / timeconstant).
    tau_t = dt / tau

    if tau_t > 0.1:
        factor = exp(-1.0 / tau_t)
    else:
        factor = 0.0

    if rng is None:
        rng = RNG

    def resample_kin(cur_kinetic_energy):
        """
        Parameters
        ----------
        cur_kinetic_energy : float
            Present value of the kinetic energy of the atoms to be thermalized
            in arbitrary units.
        """
        rr = rng.normal()
        new_kinetic_energy = (
            cur_kinetic_energy
            + (1.0 - factor)
            * (sigma * (sum_noises(dof-1) + rr**2) / dof - cur_kinetic_energy)
            + 2.0 * rr * sqrt(cur_kinetic_energy * sigma / dof * (1.0 - factor) * factor)
        )
        alpha = sqrt(new_kinetic_energy / cur_kinetic_energy)
        return alpha
    return resample_kin


def csvr_closure_2(sigma, dof, dt, tau=100, rng=None):
    if rng is None:
        rng = RNG
    c = exp(-dt / tau)

    def resample_kin(cur_kinetic_energy):
        """Canonical stocastical velocity rescaling.

        See dx.doi.org/10.1016/j.cpc.2008.01.006
        """
        R = rng.normal()
        S = np.sum(rng.normal(size=dof-1)**2)
        quot = (1 - c) * sigma / (dof * cur_kinetic_energy)
        alpha =  sqrt(c + quot * (S + R**2) + 2 * R * sqrt(c*quot))
        sign = np.sign(R + sqrt(c / quot))
        return sign * alpha
    return resample_kin


def berendsen_closure(sigma, dof, dt, tau=100, rng=None):
    """ https://doi.org/10.1063/1.448118"""
    tau_t = dt / tau

    def resample_kin(cur_kinetic_energy):
      alpha = sqrt(1 + tau_t * (sigma / cur_kinetic_energy - 1))
      return alpha
    return resample_kin



# [1] https://doi.org/10.1002/jcc.26495
#     Habershon, 2021

import itertools as it

import numpy as np

from pysisyphus.helpers_pure import get_molecular_radius


class HardSphere:
    def __init__(
        self,
        geom,
        frags,
        kappa=1.0,
        permutations=False,
        frag_radii=None,
        radii_offset=0.9452,
    ):
        """Intra-Image Inter-Molecular Hard-Sphere force.

        See A.2. in [1], Eq. (A1).
        """
        self.frags = frags
        self.kappa = kappa

        self.frag_num = len(self.frags)
        self.frag_sizes = np.array([len(frag) for frag in self.frags])
        self.pair_inds = np.array(list(it.combinations(range(self.frag_num), 2)))
        it_func = it.permutations if permutations else it.combinations
        self.pair_inds = np.array(list(it_func(range(self.frag_num), 2)))
        self.frag_inds = np.array([m for m, n in self.pair_inds])

        c3d = geom.coords3d
        frag_c3ds = [c3d[frag] for frag in self.frags]
        self.frag_radii = frag_radii
        if self.frag_radii is None:
            self.frag_radii = [
                get_molecular_radius(frag_c3d, min_offset=radii_offset)
                for frag_c3d in frag_c3ds
            ]
        self.radii_sums = np.array(
            [self.frag_radii[i] + self.frag_radii[j] for i, j in self.pair_inds]
        )

    def get_forces(self, atoms, coords, kappa=None):
        if kappa is None:
            kappa = self.kappa
        c3d = coords.reshape(-1, 3)

        # Break early when only 1 fragment is present
        if len(self.pair_inds) == 0:
            return {"energy": 1, "forces": np.zeros_like(coords)}

        centroids = np.array([c3d[frag].mean(axis=0) for frag in self.frags])
        mm, nn = np.array(self.pair_inds).T
        gdiffs = centroids[mm] - centroids[nn]
        # Add small number to avoid division by zero in 'frag_gradient' calculation
        gnorms = np.linalg.norm(gdiffs, axis=1) + 1e-16
        H = (gnorms < self.radii_sums).astype(int)
        N = H.copy()
        N *= 3 * self.frag_sizes[self.frag_inds]
        N_invs = np.divide(1, N, out=np.zeros_like(N).astype(float), where=N != 0)
        phi = kappa * N_invs * (gnorms - self.radii_sums)
        frag_gradient = (phi * H / gnorms)[:, None] * gdiffs
        gradient = np.zeros_like(c3d)
        # Distribute gradient onto fragments
        for frag_ind, ff in zip(self.frag_inds, frag_gradient):
            gradient[self.frags[frag_ind]] += ff
        forces = -gradient.flatten()

        f3d = forces.reshape(-1, 3)
        f3d -= f3d.mean(axis=0)[None, :]

        return {"energy": 1, "forces": forces}


class PWHardSphere:
    def __init__(
        self,
        geom,
        frags,
        sub_frags,
        kappa=1.0,
    ):
        """Inter-Molecular pairwise Hard-Sphere forces between atoms.

        Hardsphere forces are only applied between certain atoms of given fragments,
        but the whole fragment is moved. Can be used to remove atom inter-molecular
        atom clashes.
        """
        self.frags = frags
        self.sub_frags = sub_frags
        self.kappa = kappa

        self.frag_num = len(self.frags)
        self.frag_sizes = np.array([len(frag) for frag in self.frags])
        self.pair_inds = np.array(list(it.combinations(range(self.frag_num), 2)))
        cov_rads = geom.covalent_radii
        self.pair_atom_inds = list()
        self.pair_cov_radii = list()
        for m, n in self.pair_inds:
            frag_m = self.sub_frags[m]
            frag_n = self.sub_frags[n]
            painds = list()
            pcr = list()
            for fm in frag_m:
                crm = cov_rads[m]
                for fn in frag_n:
                    crn = cov_rads[m]
                    crsum = crm + crn
                    painds.append([fm, fn])
                    pcr.append(crsum)
            self.pair_atom_inds.append(painds)
            self.pair_cov_radii.append(pcr)

    def get_forces(self, atoms, coords, kappa=None):
        if kappa is None:
            kappa = self.kappa
        c3d = coords.reshape(-1, 3)

        # Break early when only 1 fragment is present
        if len(self.pair_inds) == 0:
            return {"energy": 1, "forces": np.zeros_like(coords)}

        forces = np.zeros_like(c3d)
        centroids = np.array([c3d[frag].mean(axis=0) for frag in self.frags])
        N = 1.0
        N_inv = 1 / N
        for (m, n), pai, pcr in zip(
            self.pair_inds, self.pair_atom_inds, self.pair_cov_radii
        ):
            frag_m = self.frags[m]
            frag_n = self.frags[n]
            centr_m = centroids[m]
            centr_n = centroids[n]
            dcentr = centr_m - centr_n
            force_dir = dcentr / np.linalg.norm(dcentr)
            for (am, an), crmn in zip(pai, pcr):
                amn = c3d[am] - c3d[an]
                distmn = np.linalg.norm(amn)
                diff = distmn - crmn
                fact = int(distmn < crmn)
                if not fact:
                    continue
                # Magnitude of applied force
                magn = kappa * N_inv * diff
                force = magn * force_dir
                # Distribute half of the force onto each fragment
                force_2 = force / 2
                # The signs below depend on the difference centr_m - centr_n above
                forces[frag_m] -= force_2
                forces[frag_n] += force_2

        forces -= forces.mean(axis=0)[None, :]
        forces = forces.flatten()

        return {"energy": 1, "forces": forces}



# [1] https://www.sciencedirect.com/science/article/pii/S0166128098004758
#     https://doi.org/10.1016/S0166-1280(98)00475-8
#     Dapprich, Frisch, 1998
# [2] https://onlinelibrary.wiley.com/doi/abs/10.1002/9783527629213.ch2
#     Clemente, Frisch, 2010
#
# Not implemented in pysisyphus
#
# [2] https://aip.scitation.org/doi/pdf/10.1063/1.2814164?class=pdf
#     QM/QM ONIOM EE based on Mulliken charges
#     Hratchian, Raghavachari, 2008
# [3] https://aip.scitation.org/doi/full/10.1063/1.3315417<Paste>
#     QM/QM ONIOM EE based on Löwdin charges
#     Mayhall, Hratchian, 2010
# [4] https://www.frontiersin.org/articles/10.3389/fchem.2018.00089/full
#     Overview on hybrid methods
# [5] https://doi.org/10.1021/jp0446332
#     Electronic embedding charge redistribution
#     Lin, Truhlar 2004
# Excited state ONIOM
# [5] https://aip.scitation.org/doi/pdf/10.1063/1.4972000?class=pdf
# [6] https://pubs.rsc.org/en/content/articlehtml/2012/pc/c2pc90007f


import itertools as it
import logging
from collections import namedtuple

import numpy as np
import scipy.sparse as sparse

from pysisyphus.calculators import (
    Composite,
    Gaussian16,
    MOPAC,
    OBabel,
    OpenMolcas,
    ORCA,
    ORCA5,
    Psi4,
    Turbomole,
    XTB,
    PyXTB,
)
from pysisyphus.calculators.Calculator import Calculator
from pysisyphus.elem_data import COVALENT_RADII as CR
from pysisyphus.Geometry import Geometry
from pysisyphus.helpers_pure import full_expand
from pysisyphus.intcoords.setup import get_bond_sets
from pysisyphus.intcoords.setup_fast import get_bond_vec_getter


CALC_DICT = {
    "composite": Composite,
    "g16": Gaussian16,
    "openmolcas": OpenMolcas.OpenMolcas,
    "mopac": MOPAC,
    "obabel": OBabel,
    "orca": ORCA.ORCA,
    "orca5": ORCA5.ORCA5,
    "psi4": Psi4,
    "turbomole": Turbomole,
    "xtb": XTB.XTB,
    # "pypsi4": PyPsi4,
    "pyxtb": PyXTB.PyXTB,
}
try:
    from pysisyphus.calculators.PySCF import PySCF

    CALC_DICT["pyscf"] = PySCF
except (ModuleNotFoundError, ImportError, OSError):
    # print("Error importing PySCF in ONIOMv2")
    pass


Link = namedtuple("Link", "ind parent_ind atom g")


def get_g_value(atom, parent_atom, link_atom):
    cr, pcr, lcr = [CR[a.lower()] for a in (atom, parent_atom, link_atom)]

    # Ratio between sum of CR_atom and CR_link with sum of CR_atom CR_parent_atom.
    # See [1] Sect. 2.2 page 5.
    g = (cr + lcr) / (cr + pcr)
    return g


def cap_fragment(atoms, coords, fragment, link_atom="H", g=None):
    coords3d = coords.reshape(-1, 3)

    fragment_set = set(fragment)

    # Determine bond(s) that connect fragment with the rest
    bonds = get_bond_sets(atoms, coords3d)
    bond_sets = [set(b) for b in bonds]

    # Find all bonds that involve one atom of model. These bonds
    # connect the model to the real geometry. We want to cap these
    # bonds.
    break_bonds = [b for b in bond_sets if len(b & fragment_set) == 1]

    # Put capping atoms at every bond to break.
    # The model fragment size corresponds to the length of the union of
    # the model set and the atoms in break_bonds.
    capped_frag = fragment_set.union(*break_bonds)
    capped_inds = list(sorted(capped_frag))

    # Index map between the new model geometry and the original indices
    # in the real geometry.
    atom_map = {
        model_ind: real_ind
        for real_ind, model_ind in zip(capped_inds, range(len(capped_inds)))
    }

    links = list()
    for bb in break_bonds:
        to_cap = bb - fragment_set
        assert len(to_cap) == 1
        ind = list(bb - to_cap)[0]
        parent_ind = tuple(to_cap)[0]
        if g is None:
            g = get_g_value(atoms[ind], atoms[parent_ind], link_atom)
        link = Link(ind=ind, parent_ind=parent_ind, atom=link_atom, g=g)
        links.append(link)

    return atom_map, links


def atom_inds_to_cart_inds(atom_inds):
    stencil = np.array((0, 1, 2), dtype=int)
    size_ = len(atom_inds)
    cart_inds = np.tile(stencil, size_) + np.repeat(atom_inds, 3) * 3
    return cart_inds


class ModelDummyCalc:
    def __init__(self, model, cap=False):  # , all_atoms, all_coords):
        self.model = model
        self.cap = cap

    def get_energy(self, atoms, coords):
        energy = self.model.get_energy(atoms, coords, cap=self.cap)
        results = {"energy": energy}
        return results

    def get_forces(self, atoms, coords):
        # if self.parent_name is not None:
        # raise Exception("Currently, this does not work for Models with a parent!")
        energy, forces = self.model.get_forces(atoms, coords, cap=self.cap)
        forces_ = np.zeros((len(atoms), 3))
        # The redistribution below only works withouth parent, as otherwise
        # the numer of atoms on the RHS and LHS differ.
        forces_[: len(atoms) - len(self.model.links)] = forces.reshape(-1, 3)[
            self.model.atom_inds
        ]
        results = {"energy": energy, "forces": forces_.flatten()}
        return results

    # def get_hessian(self, atoms, coords):
    # energy, hessian = self.model.get_hessian(atoms, coords, cap=False)
    # results = {"energy": energy, "hessian": hessian}
    # return results


class Model:
    def __init__(
        self,
        name,
        calc_level,
        calc,
        parent_name,
        parent_calc_level,
        parent_calc,
        atom_inds,
        parent_atom_inds,
        use_link_atoms=True,
    ):

        self.name = name
        self.calc_level = calc_level
        self.calc = calc

        self.parent_name = parent_name
        self.parent_calc_level = parent_calc_level
        self.parent_calc = parent_calc

        self.atom_inds = list(atom_inds)
        # parent_atom_inds may be None
        try:
            self.parent_atom_inds = list(parent_atom_inds)
        except TypeError:
            self.parent_atom_inds = None

        self.use_link_atoms = use_link_atoms

        self.links = list()
        self.capped = False
        self.J = None

        self.log(f"Created model '{self}' with {len(self.atom_inds)} atoms.")

    def log(self, message=""):
        logger = logging.getLogger("calculator")
        logger.debug(self.__str__() + " " + message)

    def create_links(self, atoms, coords, debug=False):
        self.capped = True

        if self.use_link_atoms and self.parent_name is not None:
            _, self.links = cap_fragment(atoms, coords, self.atom_inds)
        self.capped_atom_num = len(self.atom_inds) + len(self.links)
        for i, link in enumerate(self.links):
            ind, parent_ind = link.ind, link.parent_ind
            self.log(
                f"\tCreated Link atom ({link.atom}) between {atoms[ind]}{ind} "
                f"and {atoms[parent_ind]}{parent_ind} (g={link.g:.6f})"
            )

        if len(self.links) == 0:
            self.log("Didn't create any link atoms!\n")

        # self.J = self.get_jacobian()
        self.J = self.get_sparse_jacobian()

        if debug:
            catoms, ccoords = self.capped_atoms_coords(atoms, coords)
            geom = Geometry(catoms, ccoords)
            geom.jmol()

    def capped_atoms_coords(self, all_atoms, all_coords):
        assert self.capped, "Did you forget to call create_links()?"

        org_atom_num = len(self.atom_inds)
        c3d = all_coords.reshape(-1, 3)

        capped_atoms = [all_atoms[i] for i in self.atom_inds]
        # Initialize empty coordinate array
        capped_coords = np.zeros((self.capped_atom_num, 3))
        # Copy non-capped coordinates
        capped_coords[:org_atom_num] = c3d[self.atom_inds]

        for i, link in enumerate(self.links):
            capped_atoms.append(link.atom)

            r1 = c3d[link.ind]
            r3 = c3d[link.parent_ind]
            r2 = r1 + link.g * (r3 - r1)
            capped_coords[org_atom_num + i] = r2
        return capped_atoms, capped_coords

    def create_bond_vec_getters(self, atoms):
        link_parent_inds = [link.parent_ind for link in self.links]
        no_bonds_with = [
            [
                link.ind,
            ]
            for link in self.links
        ]
        self.log(
            f"Model has {len(link_parent_inds)} link atom hosts: {link_parent_inds}"
        )
        covalent_radii = [CR[atom.lower()] for atom in atoms]
        self.get_bond_vecs = get_bond_vec_getter(
            atoms,
            covalent_radii,
            link_parent_inds,
            no_bonds_with,
        )

    def get_jacobian(self):
        try:
            # Shape of Jacobian is (model + link, real). TypeError will be raised
            # when self.parent_atom_inds is None.
            jac_shape = (
                len(self.atom_inds) * 3 + len(self.links) * 3,
                len(self.parent_atom_inds) * 3,
            )
        except TypeError:
            return None

        J = np.zeros(jac_shape)
        # Stencil for diagonal elements of 3x3 submatrix
        stencil = np.array((0, 1, 2), dtype=int)
        size_ = len(self.atom_inds)

        model_rows = np.arange(size_ * 3)
        # When more than two layers are present the inner layers aren't directly
        # embedded in the outermost layer. This means parent_inds does not begin
        # at 0, but with a higher index. So we need a map of the actual indices
        # (not starting at 0) to the indices in the Jacobian which start at 0.
        atom_inds = [self.parent_atom_inds.index(ind) for ind in self.atom_inds]
        ind_map = {k: v for k, v in zip(self.atom_inds, atom_inds)}
        model_cols = atom_inds_to_cart_inds(atom_inds)
        J[model_rows, model_cols] = 1

        # Link atoms
        link_start = model_rows.max() + 1
        for i, (ind, parent_ind, atom, g) in enumerate(self.links):
            rows = link_start + i * 3 + stencil
            cols = ind_map[ind] * 3 + stencil

            J[rows, cols] = 1 - g
            try:
                parent_cols = self.parent_atom_inds.index(parent_ind) * 3 + stencil
                J[rows, parent_cols] = g
            # Raised when link atom is not coupled to layer above, but
            # to a layer higher above.
            except ValueError:
                pass

        return J

    def get_sparse_jacobian(self):
        try:
            # Shape of Jacobian is (model + link, real). TypeError will be raised
            # when self.parent_atom_inds is None.
            jac_shape = (
                len(self.atom_inds) * 3 + len(self.links) * 3,
                len(self.parent_atom_inds) * 3,
            )
        except TypeError:
            return None

        # Stencil for diagonal elements of 3x3 submatrix
        stencil = np.array((0, 1, 2), dtype=int)
        ones = np.ones_like(stencil)
        size_ = len(self.atom_inds)

        model_rows = np.arange(size_ * 3)
        # When more than two layers are present the inner layers aren't directly
        # embedded in the outermost layer. This means parent_inds does not begin
        # at 0, but with a higher index. So we need a map of the actual indices
        # (not starting at 0) to the indices in the Jacobian which start at 0.
        atom_inds = [self.parent_atom_inds.index(ind) for ind in self.atom_inds]
        ind_map = {k: v for k, v in zip(self.atom_inds, atom_inds)}
        model_cols = atom_inds_to_cart_inds(atom_inds)

        jac_rows = model_rows.tolist()
        jac_cols = model_cols.tolist()
        jac_data = np.ones_like(jac_cols).tolist()

        # Link atoms
        link_start = model_rows.max() + 1
        for i, (ind, parent_ind, atom, g) in enumerate(self.links):
            rows = (link_start + i * 3 + stencil).tolist()
            cols = (ind_map[ind] * 3 + stencil).tolist()
            jac_rows += rows
            jac_cols += cols
            jac_data += (ones - g).tolist()

            try:
                parent_cols = (
                    self.parent_atom_inds.index(parent_ind) * 3 + stencil
                ).tolist()
                jac_rows += rows
                jac_cols += parent_cols
                jac_data += np.full_like(parent_cols, g, dtype=float).tolist()
            # Raised when link atom is not coupled to layer above, but
            # to a layer higher above.
            except ValueError:
                pass
        J = sparse.csr_matrix((jac_data, (jac_rows, jac_cols)), shape=jac_shape)

        return J

    def get_energy(
        self, atoms, coords, point_charges=None, parent_correction=True, cap=True
    ):
        self.log("Energy calculation")
        if cap:
            catoms, ccoords = self.capped_atoms_coords(atoms, coords)
        else:
            catoms = atoms
            ccoords = coords

        prepare_kwargs = {
            "point_charges": point_charges,
        }

        self.log("Calculation at layer level")
        results = self.calc.get_energy(catoms, ccoords, **prepare_kwargs)
        energy = results["energy"]

        # Calculate correction if parent layer is present and it is requested
        if (self.parent_calc is not None) and parent_correction:
            self.log("Calculation at parent layer level")
            parent_results = self.parent_calc.get_energy(
                catoms, ccoords, **prepare_kwargs
            )
            parent_energy = parent_results["energy"]
            energy -= parent_energy
        elif not parent_correction:
            self.log("No parent correction!")

        return energy

    def get_forces(
        self, atoms, coords, point_charges=None, parent_correction=True, cap=True
    ):
        self.log("Force calculation")
        # catoms, ccoords = self.capped_atoms_coords(atoms, coords)
        if cap:
            catoms, ccoords = self.capped_atoms_coords(atoms, coords)
        else:
            catoms = atoms
            ccoords = coords

        prepare_kwargs = {
            "point_charges": point_charges,
        }

        self.log("Calculation at layer level")
        results = self.calc.get_forces(catoms, ccoords, **prepare_kwargs)
        # These forces can contain contributions from link atoms.
        forces = results["forces"]
        energy = results["energy"]
        # Redistribute link atom forces onto the two link atom hosts using the
        # Jacobian J. Afterwards, the forces will have the shape of the parent-forces.
        if self.J is not None:
            # forces = forces.dot(self.J)
            # f^T J = (J^T f)^T
            # The transpose of the term in brackets can be ignored here, as numpy
            # does not distinguish between f and f^T for a 1d-array.
            forces = self.J.T @ forces

        # Calculate correction if parent layer is present and it is requested
        if (self.parent_calc is not None) and parent_correction:
            self.log("Calculation at parent layer level")
            parent_results = self.parent_calc.get_forces(
                catoms, ccoords, **prepare_kwargs
            )
            parent_forces = parent_results["forces"]
            parent_energy = parent_results["energy"]

            # Correct energy and forces
            energy -= parent_energy
            forces -= self.J.T @ parent_forces
        elif not parent_correction:
            self.log("No parent correction!")

        return energy, forces

    def get_hessian(
        self, atoms, coords, point_charges=None, parent_correction=True, cap=True
    ):
        self.log("Hessian calculation")
        # catoms, ccoords = self.capped_atoms_coords(atoms, coords)
        if cap:
            catoms, ccoords = self.capped_atoms_coords(atoms, coords)
        else:
            catoms = atoms
            ccoords = coords

        prepare_kwargs = {
            "point_charges": point_charges,
        }

        self.log("Calculation at layer level")
        # results = self.calc.get_hessian(catoms, ccoords, prepare_kwargs)
        results = self.calc.get_hessian(catoms, ccoords, **prepare_kwargs)
        hessian = results["hessian"]
        energy = results["energy"]
        if self.J is not None:
            # hessian = self.J.T.dot(hessian.dot(self.J))
            hessian = (self.J.T @ hessian) @ self.J

        # Calculate correction if parent layer is present and it is requested
        if (self.parent_calc is not None) and parent_correction:
            self.log("Calculation at parent layer level")
            parent_results = self.parent_calc.get_hessian(
                catoms, ccoords, **prepare_kwargs
            )
            parent_hessian = parent_results["hessian"]
            parent_energy = parent_results["energy"]

            # Correct energy and hessian
            energy -= parent_energy
            # hessian -= self.J.T.dot(parent_hessian.dot(self.J))
            hessian -= (self.J.T @ parent_hessian) @ self.J
        elif not parent_correction:
            self.log("No parent correction!")

        return energy, hessian

    # def get_delta_S(self, atoms, coords):
    # self.log("ΔS calculation")
    # catoms, ccoords = self.capped_atoms_coords(atoms, coords)

    # # Parent calculator
    # E_parent_real = self.parent_calc.get_energy(atoms, coords)["energy"]
    # self.parent_calc.reset()
    # E_parent_model = self.parent_calc.get_energy(catoms, ccoords)["energy"]
    # S_low = E_parent_real - E_parent_model
    # self.log(f"S_low={S_low:.6f} au")
    # print(f"S_low={S_low:.6f} au")
    # # High level calculator
    # E_high_real = self.calc.get_energy(atoms, coords)["energy"]
    # self.calc.reset()
    # E_high_model = self.calc.get_energy(catoms, ccoords)["energy"]
    # S_high = E_high_real - E_high_model
    # self.log(f"S_high={S_high:.6f} au")
    # print(f"S_high={S_high:.6f} au")

    # delta_S = S_low - S_high
    # self.log(f"ΔS={delta_S:.6f} au")
    # print(f"ΔS={delta_S:.6f} au")

    # return delta_S

    def parse_charges(self):
        charges = self.calc.parse_charges()
        try:
            parent_charges = self.parent_calc.parse_charges()
        except AttributeError:
            parent_charges = None

        return charges, parent_charges

    def as_geom(self, all_atoms, all_coords):
        capped_atoms, capped_coords3d = self.capped_atoms_coords(all_atoms, all_coords)
        geom = Geometry(capped_atoms, capped_coords3d)
        dummy_calc = self.as_calculator()
        geom.set_calculator(dummy_calc)
        return geom

    def as_calculator(self, cap=False):
        return ModelDummyCalc(self, cap=cap)

    def __str__(self):
        # return (
        # f"Model({self.name}, {len(self.atom_inds)} atoms, "
        # f"level={self.calc_level}, parent_level={self.parent_calc_level})"
        # )
        return f"{self.name}_{self.calc_level}"

    def __repr__(self):
        return self.__str__()


def get_embedding_charges(embedding, layer, parent_layer, coords3d):
    # Only consider charges that belong to atoms in the parent
    # layer. Otherwise this would result in additonal charges at
    # the same positions as the atoms we would like to calculate.
    if "electronic" in embedding:
        assert (
            len(parent_layer) == 1
        ), "Multicenter ONIOM in intermediate layer is not supported!"
        parent_model = parent_layer[0]
        parent_inds = parent_model.atom_inds
        point_charges, _ = parent_model.parse_charges()

        layer_inds = set(*it.chain([model.atom_inds for model in layer]))
        # Determine indices of atoms that are in the parent layer, but
        # not in the current layer
        only_parent_inds = list(set(parent_inds) - layer_inds)

    del_charge_inds = list()
    all_redist_coords_charges = list()
    # Here, redistributed and scaled charges are calculated. In the EE-RC and EE-RCD
    # schemes the link atom parent (LAP) charges are divided by the number of bonds
    # connected to the LAP minus 1. They are put halfway along theses bonds.
    # See [5] for a discussion.
    # EE-RC and EE-RCD are very similar. The block below handles calculations that
    # are common to both methods, e.g. calculation of the redistributed charges and
    # their coordinates.
    #
    # This will be executed for 'electronic_rc' and 'electronic_rcd'
    if "electronic_rc" in embedding:
        # Collect charges for models in a layer, e.g., for multicenter ONIOM.
        for model in layer:
            redist_coords_charges = list()
            single_redist_charges = list()
            # Determine bonds, connected to link parent.
            link_host_bond_vecs, bonded_inds = model.get_bond_vecs(
                coords3d, return_bonded_inds=True
            )
            # Determine link atoms
            links = model.links
            for link, bond_vecs in zip(links, link_host_bond_vecs):
                parent_ind = link.parent_ind
                # Presence of a link atom implies a bond.
                assert len(bond_vecs) > 0
                # *parent_coords, parent_charge = point_charges[link.parent_ind]
                # parent_charge = ee_charges[parent_ind]
                parent_charge = point_charges[parent_ind]
                parent_coords = coords3d[parent_ind]
                bond_num = len(bond_vecs)
                redist_charge = parent_charge / bond_num
                single_redist_charges.append(redist_charge)
                # Put modified charges halfway on the bonds
                redist_coords = parent_coords + bond_vecs / 2
                redist_coords_charges.extend(
                    [(*coords, redist_charge) for coords in redist_coords]
                )
                del_charge_inds.append(parent_ind)
            redist_coords_charges = np.array(redist_coords_charges)

            # Redistributed charges and dipoles to preserve the M1-M2 bond dipoles. See [5].
            if embedding == "electronic_rcd":
                # Multiply all redistributed charges by 2
                redist_coords_charges[:, -1] *= 2
                # Substract original redistributed charge from M2 charges
                for binds, src in zip(bonded_inds, single_redist_charges):
                    point_charges[binds] -= src

            # Gather redistributed charges of separate models (centers)
            all_redist_coords_charges.extend(redist_coords_charges)

    assert len(del_charge_inds) == len(set(del_charge_inds)), (
        "It seems that one parent hosts multiple link atoms. I did not think about "
        "cases like that yet!"
    )
    # Only keep charges that are not on link atom hosts/parents
    keep_mask = [opi for opi in only_parent_inds if opi not in del_charge_inds]
    kept_point_charges = point_charges[keep_mask]
    kept_coords3d = coords3d[keep_mask]
    kept_coords_point_charges = np.concatenate(
        (kept_coords3d, kept_point_charges[:, None]), axis=1
    )

    # Join unmodified charges and redistributed charges
    if len(all_redist_coords_charges) > 0:
        kept_coords_point_charges = np.concatenate(
            (kept_coords_point_charges, all_redist_coords_charges), axis=0
        )
    return kept_coords_point_charges


class LayerCalc:
    def __init__(self, models, total_size, parent_layer_calc=None):
        self.models = models
        self.parent_layer_calc = parent_layer_calc
        self.total_size = total_size  # atoms in the total system

        self.models_str = ", ".join([str(model) for model in self.models])

    @property
    def mult(self):
        mults = [model.calc.mult for model in self.models]
        mult0 = mults[0]
        assert all([mult == mult0 for mult in mults])
        return mult0

    @mult.setter
    def mult(self, mult):
        for model in self.models:
            model.calc.mult = mult

    @property
    def charge(self):
        charges = [model.calc.mult for model in self.models]
        charge0 = charges[0]
        assert all([charge == charge0 for charge in charges])
        return charge0

    @charge.setter
    def charge(self, charge):
        for model in self.models:
            model.calc.charge = charge

    def run_calculations(self, atoms, coords, method):
        results = [getattr(model, method)(atoms, coords) for model in self.models]
        return results

    @staticmethod
    def energy_from_results(model_energies, parent_energy=None):
        energy = sum(model_energies)
        if parent_energy is not None:
            energy += parent_energy
        return energy

    def do_parent(self, with_parent):
        return (self.parent_layer_calc is not None) and with_parent

    def get_energy(self, atoms, coords, with_parent=True):
        model_energies = self.run_calculations(atoms, coords, "get_energy")
        if self.do_parent(with_parent):
            parent_result = self.parent_layer_calc.get_energy(
                atoms, coords, with_parent=True
            )
            parent_energy = parent_result["energy"]
        else:
            parent_energy = None
        energy = self.energy_from_results(model_energies, parent_energy)
        return {"energy": energy}

    def get_forces(self, atoms, coords, with_parent=True):
        full_forces = np.zeros((self.total_size, 3))
        model_energies = list()
        for model in self.models:
            model_energy, model_forces = model.get_forces(
                atoms, coords, parent_correction=True
            )
            model_energies.append(model_energy)
            full_forces[model.parent_atom_inds] = model_forces.reshape(-1, 3)
        forces = full_forces.flatten()
        if self.do_parent(with_parent):
            parent_result = self.parent_layer_calc.get_forces(
                atoms, coords, with_parent=True
            )
            parent_energy = parent_result["energy"]
            parent_forces = parent_result["forces"]
            forces += parent_forces
        else:
            parent_energy = None
        energy = self.energy_from_results(model_energies, parent_energy)
        results = {
            "energy": energy,
            "forces": forces,
        }
        return results

    def get_hessian(self, atoms, coords, with_parent=True):
        hessian = np.zeros((self.total_size*3, self.total_size*3))
        model_energies = list()
        for model in self.models:
            model_energy, model_hessian = model.get_hessian(
                    atoms, coords, parent_correction=True
            )
            model_energies.append(model_energy)
            try:
                inds = 3*model.parent_atom_inds[:, None] + np.arange(3)
                hessian[inds, inds] += model_hessian
            except TypeError:
                hessian += model_hessian
        if self.do_parent(with_parent):
            parent_result = self.parent_layer_calc.get_hessian(
                atoms, coords, with_parent=True
            )
            parent_energy = parent_result["energy"]
            parent_hessian = parent_result["hessian"]
            hessian += parent_hessian
        else:
            parent_energy = None
        energy = self.energy_from_results(model_energies, parent_energy)
        result = {
            "energy": energy,
            "hessian": hessian,
        }
        return result

    def __str__(self):
        return f"LayerCalc({self.models_str})"

    def __repr__(self):
        return self.__str__()


class ONIOM(Calculator):
    embeddings = {
        "": "",
        "electronic": "Electronic embedding",
        "electronic_rc": "Electronic embedding with redistributed charges",
        "electronic_rcd": "Electronic embedding with redistributed charges and dipoles",
    }

    def __init__(
        self,
        calcs,
        models,
        geom,
        layers=None,
        embedding="",
        real_key="real",
        use_link_atoms=True,
        *args,
        **kwargs,
    ):
        """
        layer: list of models
            len(layer) == 1: normal ONIOM, len(layer) >= 1: multicenter ONIOM.
        model:
            (sub)set of all atoms that resides in a certain layer and has
            a certain calculator.
        """

        super().__init__(*args, **kwargs)

        if embedding is None:
            embedding = ""
        assert (
            embedding in self.embeddings.keys()
        ), f"Valid embeddings are: {self.embeddings.keys()}"
        self.embedding = embedding

        assert real_key not in models, f'"{real_key}" must not be defined in "models"!'
        assert real_key in calcs, f'"{real_key}" must be defined in "calcs"!'

        self.use_link_atoms = use_link_atoms

        # Expand index-lists in models
        for model in models.values():
            if ".." in model["inds"]:
                model["inds"] = full_expand(model["inds"])

        # When no ordering of layers is given we try to guess it from
        # the size of the respective models. It's probably a better idea
        # to always specify the layer ordering though ;)
        if layers is None:
            self.log(
                "No explicit layer ordering specified! Determining layer "
                "hierarchy from model sizes. This does not support multi-"
                "center ONIOM!"
            )
            as_list = [(key, val) for key, val in models.items()]
            # Determine hierarchy of models, from biggest to smallest model
            layers = [
                key
                for key, val in sorted(
                    as_list, key=lambda model: -len(model[1]["inds"])
                )
            ]

        assert real_key not in layers, f'"{real_key}" must not be defined in "layers"!'

        ############
        #          #
        #  LAYERS  #
        #          #
        ############

        # Add real model and layer as they are missing right now. The real
        # layer is always the last layer. The real layer is always calculated
        # by the 'realkey'-calculator.
        layers = [real_key] + layers
        models[real_key] = {
            "calc": real_key,
            "inds": list(range(len(geom.atoms))),
        }
        self.log(f"Layer-ordering from big to small: {layers}")

        # Single-model layers will be given as strings. As we also support
        # multicenter-ONIOM there may also be layers that are given as lists
        # that contain multiple models per layer.
        # Now we convert the single-model layers to lists of length 1, so
        # every layer is a list.
        layers = [
            [
                layer,
            ]
            if isinstance(layer, str)
            else layer
            for layer in layers
        ]
        self.layer_num = len(layers)
        assert self.layer_num > 1, "ONIOM with only 1 layer requested. Aborting!"

        ############
        #          #
        #  MODELS  #
        #          #
        ############

        # Create mapping between model and its parent layer. Actually
        # this is a bit hacky right now, as the mapping should not be between
        # model and parent layer, but between model and parent model.
        # This way we expect the parent layer to have the same calculator
        # throughout, so multicenter ONIOM with different calculators
        # in all but the smallest layer (highest level) is not well defined.
        #
        # If multicenter ONIOM in an intermediate layer is useful may
        # be another question to be answered ;).
        self.model_parent_layers = dict()
        for i, layer in enumerate(layers[1:]):
            self.model_parent_layers.update({model: i for model in layer})
        model_keys = list(it.chain(*layers))

        cur_calc_num = 0

        def get_calc(calc_key, base_name=None):
            """Helper function for easier generation of calculators
            with incrementing calc_number."""
            nonlocal cur_calc_num

            kwargs = calcs[calc_key].copy()
            type_ = kwargs.pop("type")

            kwargs["calc_number"] = cur_calc_num
            if base_name is not None:
                kwargs["base_name"] = base_name

            calc = CALC_DICT[type_](**kwargs)
            cur_calc_num += 1
            return calc

        # Create models and required calculators.
        self.models = list()
        self.layers = [list() for _ in layers]
        for model_key in model_keys:
            model_calc_key = models[model_key]["calc"]
            model_base_name = f"{model_key}_{model_calc_key}"
            model_calc = get_calc(model_calc_key, base_name=model_base_name)
            # Update parent information
            try:
                parent_layer_ind = self.model_parent_layers[model_key]
                parent_layer = layers[parent_layer_ind]
                parent_calc_keys = set(
                    [models[model_key]["calc"] for model_key in parent_layer]
                )
                assert len(parent_calc_keys) == 1, (
                    "It seems you are trying to run a multicenter ONIOM setup in "
                    "an intermediate layer with different calculators. This is "
                    "not supported right now."
                )
                parent_name = parent_layer[0]
                parent_calc_key = models[parent_name]["calc"]
                parent_base_name = f"{model}_parent"
                parent_calc = get_calc(parent_calc_key, base_name=parent_base_name)
                parent_atom_inds = models[parent_name]["inds"]
            except KeyError:
                parent_name = parent_calc_key = parent_calc = parent_atom_inds = None
                parent_layer_ind = -1

            model = Model(
                name=model_key,
                calc_level=model_calc_key,
                calc=model_calc,
                atom_inds=models[model_key]["inds"],
                use_link_atoms=self.use_link_atoms,
                #
                parent_name=parent_name,
                parent_calc_level=parent_calc_key,
                parent_calc=parent_calc,
                parent_atom_inds=parent_atom_inds,
            )
            self.models.append(model)
            self.layers[parent_layer_ind + 1].append(model)

        self.log("Created all ONIOM layers:")
        for model in self.models:
            self.log("\t" + str(model))

        # Create link atoms
        [model.create_links(geom.atoms, geom.cart_coords) for model in self.models]
        # Create functions to calculate bond vectors with link atom hosts
        [model.create_bond_vec_getters(geom.atoms) for model in self.models]

        # And do a quick sanity check
        assert (
            len(self.models[0].links) == 0
        ), "There must not be any links in the 'real' layer!"
        # Look for link atoms that appear in two adjacent layers. In such situations
        # the higher layer is coupled to a layer two levels below. This may be a bad
        # idea.
        for i, (lower_model, model) in enumerate(
            zip(self.models[:-1], self.models[1:])
        ):
            lower_links = lower_model.links
            links = model.links
            same_links = [link for link in links if link in lower_links]
            if same_links:
                print(f"Found {len(same_links)} link(s) that appear(s) in two layers!")
                for j, link in enumerate(same_links):
                    print(f"\t{j:02d}: {link}")
                print(
                    f"Your current setup couples layer '{model.name}' to "
                    f"layer '{self.models[i-1].name}' two levels below! "
                    "This is probably a bad idea!"
                )

        self.log(
            f"Created ONIOM calculator with {self.layer_num} layers and "
            f"{len(self.models)} models."
        )

        self.layer_calcs = list()
        for i, layer in enumerate(self.layers):
            try:
                parent_layer_calc = self.layer_calcs[-1]
            except IndexError:
                parent_layer_calc = None
            layer_calc = LayerCalc(
                models=layer,
                total_size=len(geom.atoms),
                parent_layer_calc=parent_layer_calc,
            )
            self.layer_calcs.append(layer_calc)

    @property
    def model_iter(self):
        try:
            # A layer may contain several models
            model_iter = it.chain(*[layer for layer in self.layers])
        except AttributeError:
            model_iter = list()
        return model_iter

    @property
    def mult(self):
        mults = [model.calc.mult for model in self.model_iter]
        mult0 = mults[0]
        assert all([mult == mult0 for mult in mults])
        return mult0

    @mult.setter
    def mult(self, mult):
        for model in self.model_iter:
            model.calc.mult = mult

    @property
    def charge(self):
        charges = [model.calc.charge for model in self.model_iter]
        charge0 = charges[0]
        return charge0

    @charge.setter
    def charge(self, charge):
        for model in self.model_iter:
            model.calc.charge = charge

    def run_calculations(self, atoms, coords, method):
        self.log(f"{self.embeddings[self.embedding]} ONIOM calculation")

        all_results = list()
        for i, layer in enumerate(self.layers):
            point_charges = None
            # Calculate embedding charges, if required
            if self.embedding and (i > 0):
                parent_layer = self.layers[i - 1]
                coords3d = coords.reshape(-1, 3)
                point_charges = get_embedding_charges(
                    self.embedding, layer, parent_layer, coords3d
                )
                self.log(
                    f"Polarizing calculation in layer {i} ({layer}) by "
                    f"charges from layer {i-1} ({self.layers[i-1]})."
                )
                ee_charge_sum = point_charges[:, -1].sum()
                self.log(f"sum(charges)={ee_charge_sum:.4f}")

                # Enable for debugging
                # from pysisyphus.wrapper.jmol import render_geom_and_charges
                # if len(layer) == 1:
                # model = layer[0]
                # tmp_atoms, tmp_coords = model.capped_atoms_coords(atoms, coords)
                # render_geom_and_charges(
                # Geometry(tmp_atoms, tmp_coords), point_charges
                # )

            results = [
                getattr(model, method)(atoms, coords, point_charges=point_charges)
                for model in layer
            ]
            all_results.extend(results)

        self.calc_counter += 1
        return all_results

    def run_calculation(self, atoms, coords):
        self.log("run_calculation() called. Doing simple energy calculation!")
        return self.get_energy(atoms, coords)

    def get_energy(self, atoms, coords):
        all_energies = self.run_calculations(atoms, coords, "get_energy")

        energy = sum(all_energies)

        return {
            "energy": energy,
        }

    def get_forces(self, atoms, coords):
        all_results = self.run_calculations(atoms, coords, "get_forces")

        energies, forces_ = zip(*all_results)
        forces_ = [np.array(f).reshape(-1, 3) for f in forces_]
        energy = sum(energies)

        forces = forces_[0]  # first layer has shape of the total system
        for mdl, f in zip(self.models[1:], forces_[1:]):
            forces[mdl.parent_atom_inds] += f

        return {
            "energy": energy,
            "forces": forces.flatten(),
        }

    def get_hessian(self, atoms, coords):
        all_results = self.run_calculations(atoms, coords, "get_hessian")

        energies, hessians = zip(*all_results)
        energy = sum(energies)

        hessian = hessians[0]  # first layer has shape of the total system
        for mdl, h in zip(self.models[1:], hessians[1:]):
            inds = atom_inds_to_cart_inds(mdl.parent_atom_inds)
            # Keep in mind that we modify hessians[0] in place
            hessian[inds[:, None], inds[None, :]] += h

        return {
            "energy": energy,
            "hessian": hessian,
        }

    def atom_inds_in_layer(self, index, exclude_inner=False):
        """Returns list of atom indices in layer at index.

        Atoms that also appear in inner layer can be excluded on request.

        Parameters
        ----------
        index : int
            pasd
        exclude_inner : bool, default=False, optional
            Whether to exclude atom indices that also appear in inner layers.

        Returns
        -------
        atom_indices : list
            List containing the atom indices in the selected layer.
        """

        layer = self.layers[index]
        atom_inds = list(it.chain(*[model.atom_inds for model in layer]))
        if exclude_inner and (index < len(self.layers) - 1):
            lower_inds = self.atom_inds_in_layer(index + 1)
            # Drop indices that appear in inner layers
            atom_inds = [i for i in atom_inds if i not in lower_inds]
        return atom_inds

    def get_layer_calc(self, layer_ind):
        return self.layer_calcs[layer_ind]

    def calc_layer(self, atoms, coords, index, parent_correction=True):
        layer = self.layers[index]
        assert len(layer) == 1, "Multicenter not yet supported!"
        (model,) = layer
        result = model.get_forces(atoms, coords, parent_correction=parent_correction)
        return result



import re
from collections.abc import Mapping

import numpy as np

from thermoanalysis.constants import C, ANG2M, AMU2KG, PLANCK, KB


class QCData:
    def __init__(
        self,
        inp,
        point_group="c1",
        mult=None,
    ):
        self.point_group = point_group.lower()
        self.symmetry_number = self.get_symmetry_number()
        if not isinstance(inp, Mapping):
            raise TypeError(
                "QCData in pdb2reaction expects a mapping with keys: "
                "coords3d, wavenumbers, scf_energy, masses, mult."
            )
        data = dict(inp)

        if mult is not None:
            data["mult"] = mult

        # Actually set data
        self.set_data(data)

        self.standard_orientation()
        I = self.inertia_tensor()
        w, v = np.linalg.eigh(I)
        self._linear = (abs(w[0]) < 1e-8) and (abs(w[1] - w[2]) < 1e-8)

    def set_data(self, data):
        expect = set("coords3d wavenumbers scf_energy masses mult".split())
        present = set(data.keys())
        missing = expect - present
        assert len(missing) == 0, f"Keys '{missing}' are missing!"

        self.masses = np.array(data["masses"], dtype=float)
        self.wavenumbers = np.array(data["wavenumbers"], dtype=float)
        self.coords3d = np.array(data["coords3d"], dtype=float).reshape(-1, 3)
        assert self.coords3d.size == 3 * len(self.masses)
        self.scf_energy = float(data["scf_energy"])
        self.mult = int(data["mult"])

    @property
    def atom_num(self):
        return len(self.masses)

    @property
    def M(self):
        """Molecular mass.

        Returns
        -------
        M : float
            Total molecular mass in amu.
        """
        return self.masses.sum()

    @property
    def mult(self):
        """Multiplicity.

        Returns
        -------
        2S+1 : int
            Multiplicity.
        """
        return self._mult

    @mult.setter
    def mult(self, mult):
        self._mult = mult

    @property
    def vib_frequencies(self):
        """Vibrational frequencies.

        Returns
        -------
        vibfreqs : np.array
            Vibrational frequencies in 1/s.
        """
        return C * self.wavenumbers * 100

    @property
    def is_linear(self):
        """Wether the molecule is linear.

        Returns
        -------
        is_linear : bool
            Wether the molecule is linear.
        """
        # return self.point_group in ("cinf", "dinfh")
        return self._linear

    @property
    def is_atom(self):
        """Wether the 'molecule' consists of only an atom.

        Returns
        -------
        is_atoms : bool
            Wether the molecule is only one atom.
        """
        return len(self.masses) == 1

    @property
    def rot_temperatures(self):
        """Rotational temperatures in K.

        Returns
        -------
        rot_temps : np.array
            Rotational temperatures in K.
        """
        if self.is_atom:
            return np.full(3, np.nan)

        self.standard_orientation()
        I = self.inertia_tensor() * ANG2M**2 * AMU2KG
        w, v = np.linalg.eigh(I)
        rot_temps = PLANCK**2 / (8 * np.pi**2 * w * KB)
        return rot_temps

    def get_symmetry_number(self, point_group=None):
        """Symmetry number for rotatioanl partiton function.

        Returns
        -------
        symmetry_number : int
            Symmetry number for calculation of rotational terms.
        """
        symm_dict = {
            "c1": 1,
            "ci": 1,
            "cs": 1,
            "cinf": 1,
            "dinfh": 2,
            "t": 12,
            "td": 12,
            "oh": 24,
            "ih": 60,
        }
        if point_group is None:
            point_group = self.point_group
        pg = point_group.lower()
        try:
            return symm_dict[pg]
        except KeyError:
            pass
        regex = r"[cds](\d+)"
        mobj = re.match(regex, pg)
        try:
            sym_num = int(mobj[1])
        except TypeError:
            raise Exception(f"Specified point group '{pg}' is invalid!")

        if pg.startswith("d"):
            sym_num *= 2
        elif pg.startswith("s"):
            sym_num /= 2
        assert sym_num == int(sym_num), (
            "Check your point group! Did you "
            "specify some 'Sn' group with n ∈ (1, 3, 5, ...)? Please use "
            "the corresponding 'Cnm' groups instead!"
        )
        return sym_num

    def inertia_tensor(self):
        """Inertita tensor.

                              | x² xy xz |
        (x y z)^T . (x y z) = | xy y² yz |
                              | xz yz z² |
        Returns
        -------
        I : np.array, shape (3, 3)
            Ineratia tensor  in units of Angstrom² * amu.
        """
        x, y, z = self.coords3d.T
        squares = np.sum(self.coords3d**2 * self.masses[:, None], axis=0)
        I_xx = squares[1] + squares[2]
        I_yy = squares[0] + squares[2]
        I_zz = squares[0] + squares[1]
        I_xy = -np.sum(self.masses * x * y)
        I_xz = -np.sum(self.masses * x * z)
        I_yz = -np.sum(self.masses * y * z)
        I = np.array(((I_xx, I_xy, I_xz), (I_xy, I_yy, I_yz), (I_xz, I_yz, I_zz)))
        return I

    @property
    def average_moment_of_inertia(self):
        """Average moment of inertia in Angstrom² * amu."""
        w, _ = np.linalg.eigh(self.inertia_tensor())
        I_avg = np.mean(w)
        return I_avg

    @property
    def center_of_mass(self):
        """Returns the center of mass.

        Returns
        -------
        R : np.array, shape (3, )
            Center of mass in Angstrom.
        """
        return 1 / self.M * np.sum(self.coords3d * self.masses[:, None], axis=0)

    def principal_axes_are_aligned(self):
        """Check if the principal axes are aligned with the cartesian axes.

        Returns
        -------
        aligned : bool
            Wether the principal axes are aligned or not.
        """
        w, v = np.linalg.eigh(self.inertia_tensor())
        return np.allclose(v, np.eye(3)), v

    def align_principal_axes(self):
        """Align the principal axes to the cartesian axes.

        https://math.stackexchange.com/questions/145023
        """
        I = self.inertia_tensor()
        w, v = np.linalg.eigh(I)
        self.coords3d = v.T.dot(self.coords3d.T).T

    def standard_orientation(self):
        """Bring molecule in standard orientation."""
        # Translate center of mass to cartesian origin
        self.coords3d -= self.center_of_mass
        # Try to rotate the principal axes onto the cartesian axes
        for _ in range(5):
            self.align_principal_axes()
            aligned, vecs = self.principal_axes_are_aligned()
            if aligned:
                break



from dataclasses import dataclass
from typing import List, Tuple

import numpy as np
from numpy.typing import NDArray
import scipy as sp

from pysisyphus.wavefunction.helpers import symmetric_orthogonalization
from pysisyphus.wavefunction import Wavefunction


@dataclass
class PopAnalysis:
    # atoms: List[str]
    # coords3d: NDArray[float]
    pop_a: NDArray[float]
    pop_b: NDArray[float]
    nuc_charges: NDArray[float]

    @property
    def charges(self):
        return self.nuc_charges - self.pop_a - self.pop_b

    @property
    def tot_charge(self):
        return self.charges.sum()

    @property
    def spin_pop(self):
        return np.abs(self.pop_a - self.pop_b)

    @property
    def alpha_spin_pop(self):
        spin_pop = self.pop_a - self.pop_b
        beta_mask = spin_pop < 0.0
        spin_pop[beta_mask] = 0.0
        return spin_pop

    @property
    def beta_spin_pop(self):
        spin_pop = self.pop_a - self.pop_b
        alpha_mask = spin_pop > 0.0
        spin_pop[alpha_mask] = 0.0
        return np.abs(spin_pop)


def mulliken_charges(
    P: Tuple[NDArray[float]],
    S: NDArray[float],
    nuc_charges: NDArray[int],
    ao_centers: List[int],
) -> PopAnalysis:
    def mulliken_atom_pops(P: NDArray[float], S: NDArray[float]) -> NDArray[float]:
        ao_populations = np.einsum("ij,ji->i", P, S)
        atom_populations = np.zeros(len(nuc_charges))
        for i, center in enumerate(ao_centers):
            atom_populations[center] += ao_populations[i]
        return atom_populations

    atom_populations_a = mulliken_atom_pops(P[0], S)
    atom_populations_b = mulliken_atom_pops(P[1], S)

    pop_ana = PopAnalysis(
        pop_a=atom_populations_a,
        pop_b=atom_populations_b,
        nuc_charges=nuc_charges,
    )
    return pop_ana


def mulliken_charges_from_wf(wf: Wavefunction) -> NDArray[float]:
    return mulliken_charges(
        P=wf.P,
        S=wf.S,
        nuc_charges=wf.nuc_charges,
        ao_centers=wf.ao_centers,
    )


def make_iaos(
    C_occ: NDArray[float],
    S_org: NDArray[float],
    S_minao: NDArray[float],
    S_cross: NDArray[float],
) -> NDArray[float]:
    """Intrinsic atomic orbitals.

    [1] https://doi.org/10.1021/ct400687b
    """
    P12 = sp.linalg.solve(S_org, S_cross)  # Projector on basis1, S1⁻¹ S12
    P21 = sp.linalg.solve(S_minao, S_cross.T)  # Projector on basis2, S2⁻¹ S21

    # Depolarized MOs in basis2
    C2_depol = P12.dot(P21).dot(C_occ)
    C2_depol = symmetric_orthogonalization(C2_depol, S_org)

    O = C_occ.dot(C_occ.T).dot(S_org)
    O_ = C2_depol.dot(C2_depol.T).dot(S_org)
    iaos = (2 * O.dot(O_) - O_ - O).dot(P12) + P12
    iaos = symmetric_orthogonalization(iaos, S_org)

    return iaos


def iao_charges_from_wf(wf: Wavefunction) -> PopAnalysis:
    """IAO charges.

    Extension to ES is described here:
        https://chemistry.stackexchange.com/a/75913
    """
    S_org = wf.S
    C_occ = wf.C_occ
    minao_shells = wf.shells.from_basis("minao")
    S_minao = getattr(minao_shells, "S_cart" if wf.is_cartesian else "S_sph")
    S_cross = wf.S_with_shells(
        minao_shells
    )  # Overlap between original and MINAO  basis

    def get_iao_P(C_occ: NDArray[float]):
        iaos = make_iaos(C_occ, S_org, S_minao, S_cross)
        C_iao = iaos.T @ S_org @ C_occ  # Projection of original MOs onto IAOs
        P_iao = C_iao @ C_iao.T
        return P_iao

    # Always assume separate α C_occ and β C_occ matrices.
    P_iao = np.array([get_iao_P(c_occ) for c_occ in C_occ])

    minao_ao_centers = list(minao_shells.sph_ao_centers)
    return mulliken_charges(
        P=P_iao,
        S=np.eye(len(minao_ao_centers)),
        nuc_charges=wf.nuc_charges,
        ao_centers=minao_ao_centers,
    )



import argparse
from math import pi as PI, ceil
from pathlib import Path
import sys

from pysisyphus.constants import AMU2KG
from pysisyphus.helpers import geom_loader
from pysisyphus.helpers_pure import get_input, highlight_text
from pysisyphus.io.pdb import geom_to_pdb_str
from pysisyphus.wrapper.packmol import make_input, call_packmol

from pysisyphus.db import LEVELS, MOLECULES
from pysisyphus.db.helpers import get_path as db_get_path


AMU2G = AMU2KG * 1e3
CM2ANG = 1e-8


def parse_args(args):
    parser = argparse.ArgumentParser()

    # Solvent
    solvent_group = parser.add_mutually_exclusive_group(required=True)
    solvent_group.add_argument("--solv", help="Filename of solvent geometry.")
    solvent_group.add_argument(
        "--db", action="store_true", help="Choose from internal database."
    )

    parser.add_argument(
        "--solv_num", type=int, help="Number of solvent molecules to pack."
    )
    parser.add_argument("--solv_dens", type=float, help="Solvent density in g/cm³.")

    parser.add_argument(
        "--output", default="output.pdb", help="Filename of packed molecules."
    )

    # Solute
    parser.add_argument("--solute", default=None, help="Filename of solute geometry.")
    parser.add_argument(
        "--solute_num", type=int, default=1, help="Number of solute molecules to pack."
    )

    return parser.parse_args(args)


def as_pdb(fn):
    if not str(fn).endswith(".pdb"):
        geom = geom_loader(fn)
        pdb_str = geom_to_pdb_str(geom)
        cwd = Path(".")
        pdb_fn = cwd / Path(fn).with_suffix(".pdb").name
        with open(pdb_fn, "w") as handle:
            handle.write(pdb_str)
        print(f"Converted '{fn}' to PDB format ('{pdb_fn}')")
    return pdb_fn


def sphere_radius_from_volume(volume):
    radius = (3 / 4 * volume / PI) ** (1 / 3)
    return radius


def volume_for_density(molecule_num, mol_mass, density):
    # Convert density from g/cm³ to amu/Å³
    density_au = density / AMU2G * CM2ANG ** 3
    # The molar mass in g/mol is numerically equal to the value in AMU (dalton)
    # so we can use it as it is.
    total_mass = mol_mass * molecule_num
    # Volume in Å
    volume = total_mass / density_au

    return volume


def print_info(title, geom):
    print(title)
    print(f"\t{geom}")
    print(f"\tMolar mass: {geom.total_mass:.2f} g mol⁻¹")
    print()


def run():
    args = parse_args(sys.argv[1:])

    solute_fn = args.solute
    if solute_fn:
        solute = geom_loader(solute_fn)
        solute_num = args.solute_num
        solute_mass = solute.total_mass
        print_info("Solute", solute)
    else:
        solute = None
        solute_mass = 0.0

    solv_fn = args.solv
    if solv_fn:
        solv_dens = args.solv_dens
    # Load from internal db
    else:
        print(highlight_text("Interactive solvent selection"))
        level = get_input(LEVELS, "Level of theory", lbl_func=lambda lvl: lvl[0])
        print()
        molecule = get_input(MOLECULES, "Molecule", lbl_func=lambda mol: mol.name)
        print()

        solv_fn = db_get_path(molecule.name, level[0])
        solv_dens = molecule.density

    solv = geom_loader(solv_fn)
    solv_num = args.solv_num
    solv_mass = solv.total_mass
    print_info("Solvent", solv)

    solute_solv_mass = solute_mass + solv_num * solv_mass
    print(f"Total mass of solute(s) and solvent(s): {solute_solv_mass:.2f} amu")
    print()

    # Solvent volume
    solv_vol = volume_for_density(solv_num, solv_mass, solv_dens)
    print(f"Solvent volume: {solv_vol:>10.2f} Å³")

    # Solute volume; Use the solvent density for this calculation
    if solute:
        solute_vol = volume_for_density(solute_num, solute_mass, solv_dens)
        print(f" Solute volume: {solute_vol:>10.2f} Å³")
    else:
        solute_vol = 0.0

    total_vol = solv_vol + solute_vol
    print(f"  Total volume: {total_vol:>10.2f} Å³")
    print()

    radius = sphere_radius_from_volume(total_vol)
    print(f"     Sphere radius: {radius:>8.2f} Å")
    cradius = ceil(radius)
    print(f"Using ceil(radius): {cradius:>8.2f} Å")
    print()

    # Create solute/solvent PDBs if needed

    inp_kwargs = {
        "output_fn": args.output,
        "solvent_fn": as_pdb(solv_fn),
        "solvent_num": solv_num,
        "sphere_radius": cradius,
    }
    if solute:
        inp_kwargs.update({"solute_fn": as_pdb(solute_fn), "solute_num": solute_num})

    inp = make_input(**inp_kwargs)
    inp_fn = "packmol.inp"
    with open(inp_fn, "w") as handle:
        handle.write(inp)
    print(f"Wrote packmol input to '{inp_fn}'")

    proc = call_packmol(inp)

    log_fn = "packmol.log"
    with open(log_fn, "w") as handle:
        handle.write(proc.stdout)
    print(f"Wrote packmol ouput to '{log_fn}'")
    print()

    return_ = proc.returncode
    if return_ != 0:
        print(proc.stdout)
    else:
        print("packmol returned successfully!")

    return return_



import functools
import itertools as it
import json
from pathlib import Path
from typing import Tuple

import numpy as np

from pysisyphus.config import BASIS_LIB_DIR
from pysisyphus.elem_data import ATOMIC_NUMBERS
from pysisyphus.Geometry import Geometry
from pysisyphus.helpers_pure import file_or_str
from pysisyphus.wavefunction import Shell, Shells
from pysisyphus.wavefunction.helpers import L_MAP


def basis_from_json(name):
    basis_path = Path(name).with_suffix(".json")
    if not basis_path.is_absolute():
        basis_path = BASIS_LIB_DIR / basis_path

    with open(basis_path) as handle:
        data = json.load(handle)
    elements = data["elements"]
    return elements


def basis_from_orca_str(basis_str):
    """
    %basis
     newgto 1
     s 2
     1 0.7 0.5
     1 0.8 0.6
     p 2
     1 0.7 0.5
     1 0.8 0.6
     d 2
     1 0.7 0.5
     1 0.8 0.6
     d 3
     1 0.7 0.5
     1 0.8 0.6
     1 0.6 0.4
     end
    end
    """
    basis = dict()
    for line in basis_str.strip().split("\n"):
        line = line.strip().lower()
        # Skip comments
        if any([line.startswith(key) for key in ("#", "%basis", "end")]) or line == "":
            continue
        tokens = line.split()
        if tokens[0] == "newgto":
            atomic_num = int(tokens[1])
            basis.setdefault(atomic_num, dict())
        elif tokens[0].isalpha():
            ang_mom, nprims = tokens
            ang_mom = L_MAP[ang_mom]
            nprims = int(nprims)
            prev_counter = 0
            shell = {
                "angular_momentum": ang_mom,
                "coefficients": list(),
                "exponents": list(),
            }
            basis[atomic_num].setdefault("electron_shells", list()).append(shell)
        else:
            counter, exponent, coefficient = [float(t) for t in tokens]
            counter = int(counter)
            if coefficient <= 1e-8:
                continue
            assert counter == (prev_counter + 1)
            shell["coefficients"].append(coefficient)
            shell["exponents"].append(exponent)
    return basis


def basis_from_pyscf_str(basis_str):
    """
    # Comment1
    He    S
         13.6267000              0.1752300
          1.9993500              0.8934830
          0.3829930              0.0000000
    He    S
         13.6267000              0.0000000
          1.9993500              0.0000000
          0.3829930              1.0000000
    """
    basis = dict()
    for line in basis_str.strip().split("\n"):
        line = line.strip().lower()
        # Skip comments
        if line.startswith("#") or (line == "") or ("basis" in line) or ("end" in line):
            continue
        tokens = line.split()
        if tokens[0].isalpha():
            element, ang_mom = [t.lower() for t in tokens]
            ang_mom = L_MAP[ang_mom]
            if ang_mom == "sp":
                raise Exception("SP shells are not yet handled")
            atomic_num = ATOMIC_NUMBERS[element]
            shell = {
                "angular_momentum": ang_mom,
                "coefficients": list(),
                "exponents": list(),
            }
            basis.setdefault(atomic_num, dict()).setdefault(
                "electron_shells", list()
            ).append(shell)
        else:
            exponent, coefficient = [float(t) for t in tokens]
            if coefficient <= 1e-8:
                continue
            shell["coefficients"].append(coefficient)
            shell["exponents"].append(exponent)
    return basis


@file_or_str(".gbs", ".GBS")
def basis_from_gbs(text):
    """
    ****
    Be 0
    S 6 1.00
    1.101996057600000e-02 -3.733117104118833e-01
    3.570467226624002e-02 -9.552914830540544e-01
    1.156831381426176e-01  9.464046666860781e-02
    3.748133675820811e-01 -1.194926861090494e+00
    1.214395310965943e+00  2.762798723086632e-01
    3.934640807529656e+00 -1.847390284420840e+00
    """
    basis = dict()
    for line in text.strip().split("\n"):
        line = line.strip().lower()
        # Skip empty lines
        if (line == "") or ("*" in line):
            continue
        tokens = line.split()
        # New center identifier line
        if (tokens[0].isalpha()) and (len(tokens) == 2):
            atom, zero = tokens
            atomic_num = str(ATOMIC_NUMBERS[atom.lower()])
            assert int(zero) == 0, "Only one basis per atom type is supported!"
            basis.setdefault(atomic_num, dict())
        # New shell
        elif tokens[0].isalpha() and (len(tokens) == 3):
            ang_mom, nprims, scale_fact = tokens
            ang_mom = L_MAP[ang_mom.lower()]
            nprims = int(nprims)
            shell = {
                "angular_momentum": [
                    ang_mom,
                ],
                "coefficients": [[]],
                "exponents": list(),
            }
            basis[atomic_num].setdefault("electron_shells", list()).append(shell)
        else:
            exponent, coefficient = [float(t) for t in tokens]
            shell["coefficients"][0].append(coefficient)
            shell["exponents"].append(exponent)
    return basis


@functools.singledispatch
def shells_with_basis(
    atoms: Tuple,
    coords,
    basis=None,
    name=None,
    shells_cls=None,
    min_coeff=1e-8,
    **kwargs
):
    assert (basis is not None) or (name is not None)
    if shells_cls is None:
        shells_cls = Shells
    if name is not None:
        basis = basis_from_json(name)

    coords3d = np.reshape(coords, (len(atoms), 3))
    shells = list()
    for i, (atom, c3d) in enumerate(zip(atoms, coords3d)):
        Zs = str(ATOMIC_NUMBERS[atom.lower()])
        basis_shells = basis[Zs]["electron_shells"]
        for bshell in basis_shells:
            L = bshell["angular_momentum"]
            # Two special cases can appear:
            # SP shells where L has two instead of only one entry
            # General contractions where L has length one, but multiple sets of coefficients
            # are given.
            #
            # With SP shells, there will be as many L values (2), as there are sets
            # of contraction coefficients (2). With general contractions, there will be
            # only one L value, but multiple sets of contraction coefficients.
            #
            # Segemented contractions will always have one L value and one set of contraction
            # coefficients. Long story short: We can treat everything on an equal footing if
            # we just repeat the L value as many times as required, when only one L value is
            # present.
            if len(L) == 1:
                L = it.cycle(L)
            exponents = np.array(bshell["exponents"], dtype=float)
            # Unpack possible general contractions or SP shells
            for L, coeffs in zip(L, bshell["coefficients"]):
                # For now we filter out 0.0 contraction coefficients, as present in
                # generally contracted basis sets.
                coeffs = np.array(coeffs, dtype=float)
                valid = np.abs(coeffs) >= min_coeff
                shell = Shell(
                    L=L,
                    center=c3d,
                    coeffs=coeffs[valid],
                    exps=exponents[valid],
                    atomic_num=Zs,
                    center_ind=i,
                )
                shells.append(shell)
    shells = shells_cls(shells, **kwargs)
    return shells


@shells_with_basis.register
def _(geom: Geometry, **kwargs):
    return shells_with_basis(geom.atoms, geom.coords, **kwargs)


class Basis:
    """
    Read basis sets from files.
    Bring them in a suitable order. 1s2s2p3s3p4s3d etc.

    NOT YET USED
    """

    pass



import logging
from pathlib import Path

from distributed import Client

from pysisyphus.helpers import slugify_worker

LOGGERS = {
    "calculator": "calculator.log",
    "wfoverlap": "wfoverlap.log",
}


def get_fh_logger(name, log_fn):
    """Initialize a logger with 'name', level DEBUG and a FileHandler."""
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    if len(logger.handlers) == 0:
        fh = logging.FileHandler(log_fn, mode="w", delay=True)
        fh.setLevel(logging.DEBUG)
        # fmt_str = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        # fmt_str = "%(asctime)s - %(message)s"
        fmt_str = "%(asctime)s - %(message)s"
        datefmt = "%y-%m-%d %H:%M:%S"
        formatter = logging.Formatter(fmt_str, datefmt=datefmt)
        fh.setFormatter(formatter)
        logger.addHandler(fh)
        # Uncommented this for now as the host is already in the filename
        # logger.debug(f"Initialized logging on {platform.node()}")


def init_logging_base(dask_worker, log_path):
    """Prepare individual loggers for one dask_worker."""
    slug = slugify_worker(dask_worker.worker_address)
    for name, log_fn_base in LOGGERS.items():
        log_fn = log_path / f"{slug}_{log_fn_base}"
        get_fh_logger(name, log_fn)


def init_logging(log_dir="./", scheduler=None):
    """Prepare the logger in log_path. When called with scheduler
    loggers for every worker are prepared."""
    log_path = Path(log_dir)
    if scheduler:
        client = Client(scheduler)
        client.run(init_logging_base, log_path=log_path)
    else:
        for name, log_fn_base in LOGGERS.items():
            log_fn = log_path / log_fn_base
            get_fh_logger(name, log_fn)



import argparse
import sys

from pysisyphus import logger
from pysisyphus.calculators import XTB
from pysisyphus.calculators.OBabel import OBabel
from pysisyphus.config import LIB_DIR
from pysisyphus.Geometry import Geometry
from pysisyphus.helpers import geom_loader
from pysisyphus.helpers_pure import chunks
from pysisyphus.intcoords.setup import get_bond_sets
from pysisyphus.linalg import get_rot_mat_for_coords
from pysisyphus.optimizers.LBFGS import LBFGS


def get_bond_subgeom(geom, ind, invert=False):
    bond_inds = get_bond_sets(geom.atoms, geom.coords3d)
    # Determine bonds that are connected to atom 'ind' in 'geom'
    conn_bond_inds = [bond for bond in bond_inds if ind in bond]
    assert len(conn_bond_inds) == 1
    ind0, ind1 = conn_bond_inds[0]
    # Index of atom to which 'ind' is connected
    anchor_ind = ind0 if (ind == ind1) else ind1
    # conn_bond_ind = (ind0, ind1)
    conn_bond_ind = (ind, anchor_ind) if invert else (anchor_ind, ind)
    return geom.get_subgeom(conn_bond_ind), anchor_ind


def run_opt(geom, freeze_inds, ff="uff", use_xtb=False, charge=0, mult=1):
    def get_xtb_calc():
        return XTB(charge=charge, mult=mult, pal=2)

    if not use_xtb:
        try:
            from openbabel import pybel

            # Create pybel.Molecule/OBMol to set the missing bonds
            mol = pybel.readstring("xyz", geom.as_xyz())
            # Use modified pybel.Molecule
            calc = OBabel(mol=mol, ff=ff)
        except ImportError:
            logger.warning("Could not import openbabel. Trying fallback to GFN2-xTB.")
            calc = get_xtb_calc()
    else:
        calc = get_xtb_calc()

    frozen_geom = geom.copy()
    # Only some atoms
    frozen_geom.freeze_atoms = freeze_inds
    frozen_geom.set_calculator(calc)

    opt = LBFGS(frozen_geom, max_cycles=1000, max_step=0.5, dump=False)
    opt.run()

    return frozen_geom


def replace_atom(
    geom,
    ind,
    repl_geom,
    repl_ind,
    return_full=True,
    opt=False,
    use_xtb=False,
    charge=0,
    mult=1,
):
    """Replace atom with fragment."""

    geom = geom.copy(coord_type="cart")
    if len(repl_geom.atoms) == 1:
        atoms = list(geom.atoms)
        atoms[ind] = repl_geom.atoms[0]
        coords3d = geom.coords3d
        return Geometry(atoms, coords3d)

    repl_geom = repl_geom.copy(coord_type="cart")

    # Determine bond-subgeometries, which will be used to calculate a rotation
    # matrix.
    geom_sub, _ = get_bond_subgeom(geom, ind)
    # Call with invert=True, as we want the bonds to point in opposite directions.
    # Without invert=True, 'get_bond_subgeom' will always return with the atom order
    # (ind, anchor_ind). This would lead to overlapping fragments.
    repl_sub, repl_anchor = get_bond_subgeom(repl_geom, repl_ind, invert=True)

    # Determine step that translates 'repl_geom' so that the atom 'ind' in 'geom'
    # and 'anchor_ind' in 'repl_geom' coincide.
    # step = geom.coords3d[ind] - repl_geom.coords3d[repl_anchor]
    # repl_sub.coords3d += step[None, :]

    # Determine rotation matrix that aligns 'repl_geom' on on 'geom_bond'
    *_, rot_mat = get_rot_mat_for_coords(geom_sub.coords3d, repl_sub.coords3d)
    # Rotate whole 'repl_geom' with 'rot_mat'
    repl_c3d = repl_geom.coords3d
    repl_centroid = repl_c3d.mean(axis=0)
    repl_c3d_rot = (repl_c3d - repl_centroid[None, :]).dot(rot_mat) + repl_centroid[
        None, :
    ]
    repl_geom.coords = repl_c3d_rot
    # Translate repl_geom so that repl_anchor and ind coincide
    step = geom.coords3d[ind] - repl_c3d_rot[repl_anchor]
    repl_c3d_rot += step[None, :]
    repl_geom.coords = repl_c3d_rot

    # Create geometries without atoms that will be deleted/replaces
    geom_ = geom.get_subgeom_without([ind])
    repl_geom_ = repl_geom.get_subgeom_without([repl_ind])
    # Union
    union = geom_ + repl_geom_
    if opt:
        freeze_inds = [ind for ind, _ in enumerate(geom_.atoms)]
        union = run_opt(union, freeze_inds, use_xtb=use_xtb, charge=charge, mult=mult)
        # Update coords in 'repl_geom' with optimized coordinates
        repl_geom_.coords3d = union.coords3d[len(freeze_inds) :]

    # This is useful when multiple replacements are to be done. If multiple
    # replacements would be done sequentially on the same geometry, while
    # returning the union, would require taking into account altered atom ordering/
    # numbering. By disabling return_full we can only return the replacement fragment,
    # that can be added to the original geometry, after all replacement fragments
    # have been calculated.
    if return_full:
        return_geom = union
    else:
        return_geom = repl_geom_
    return return_geom


# Located in ../geom_library/replacements/
REPLACEMENTS = {
    "Ph": ("benzene.xyz", 8),
    "OMe": ("methanol.xyz", 4),
    "OEt": ("ethanol.xyz", 8),
    "OH": ("water.xyz", 1),
    "F": ("hf.xyz", 1),
    "Cl": ("hcl.xyz", 1),
    "Br": ("hbr.xyz", 1),
}


def normalize_replacements(replacements):
    replacements = list(replacements)
    for i, repl in enumerate(replacements):
        if isinstance(repl[1], str):
            ind, repl_key = repl
            repl_fn, repl_ind = REPLACEMENTS[repl_key]
            repl_geom = geom_loader(LIB_DIR / "replacements" / repl_fn)
            replacements[i] = (ind, repl_ind, repl_geom)
        elif len(repl) == 3:
            continue
        else:
            raise Exception("Invalid replacements!")
    return replacements


def replace_atoms(geom, replacements, opt=False, use_xtb=False, charge=0, mult=1):
    replacements = normalize_replacements(replacements)
    repl_frags = list()
    del_inds = list()
    for ind, repl_ind, repl_geom in replacements:
        repl_frag = replace_atom(geom, ind, repl_geom, repl_ind, False, opt=opt, use_xtb=use_xtb)
        repl_frags.append(repl_frag)
        del_inds.append(ind)
    org_atom_num = len(geom.atoms)
    union = geom.get_subgeom_without(del_inds)
    for frag in repl_frags:
        union += frag

    if opt:
        freeze_inds = [ind for ind in range(org_atom_num - len(del_inds))]
        union = run_opt(union, freeze_inds, use_xtb=use_xtb)

    return union


def parse_args(args):
    parser = argparse.ArgumentParser()

    parser.add_argument("geom_fn")
    parser.add_argument("replacements", nargs="+")
    parser.add_argument("--opt", action="store_true")
    parser.add_argument("--jmol", action="store_true")
    parser.add_argument("--out_fn", type=str, default="union.xyz")
    parser.add_argument("--charge", type=int, default=0)
    parser.add_argument("--mult", type=int, default=1)
    parser.add_argument("--xtb", action="store_true")

    return parser.parse_args(args)


def run():
    args = parse_args(sys.argv[1:])
    geom_fn = args.geom_fn
    replacements_ = args.replacements
    opt = args.opt
    jmol = args.jmol
    out_fn = args.out_fn
    charge = args.charge
    mult = args.mult
    xtb = args.xtb

    # Activate opt when xtb is given
    opt = opt or xtb

    assert len(replacements_) % 2 == 0
    replacements = [(int(ind), key) for ind, key in chunks(replacements_, 2)]

    geom = geom_loader(geom_fn)
    union = replace_atoms(geom, replacements, opt=opt, charge=charge, mult=mult, use_xtb=xtb)
    print(union.as_xyz())

    if union:
        with open(out_fn, "w") as handle:
            handle.write(union.as_xyz())
        print(f"Saved new geometry to '{out_fn}'.")
    if jmol:
        union.jmol()



import collections.abc
from difflib import SequenceMatcher
from enum import Enum
import itertools as it
import json
import logging
import math
from pathlib import Path
from typing import List, Tuple
import re
import uuid
import time
from typing import Optional

import numpy as np
from numpy.typing import NDArray
import psutil

from pysisyphus.config import p_DEFAULT, T_DEFAULT
from pysisyphus.constants import AU2J, BOHR2ANG, C, R, AU2KJPERMOL, NA

import torch

"""Functions defined here don't import anything from pysisyphus, besides
the constants module, but only from the stdlib and from third parties."""


HASH_PREC = 4


def eigval_to_wavenumber(ev):
    # This approach seems numerically more unstable
    # conv = AU2J / (AMU2KG * BOHR2M ** 2) / (2 * np.pi * 3e10)**2
    # w2nu = np.sign(ev) * np.sqrt(np.abs(ev) * conv)
    # The two lines below are adopted from Psi4 and seem more stable,
    # compared to the approach above.
    conv = np.sqrt(NA * AU2J * 1.0e19) / (2 * np.pi * C * BOHR2ANG)
    w2nu = np.sign(ev) * np.sqrt(np.abs(ev)) * conv
    return w2nu


def hash_arr(arr, precision=HASH_PREC):
    str_ = np.array2string(arr, precision=precision)
    return hash(str_)


def hash_args(*args, precision=HASH_PREC):
    hashes = list()
    for arg in args:
        try:
            hash_ = hash(arg)
        except TypeError:
            hash_ = hash_arr(arg, precision=precision)
        hashes.append(hash_)
    return hash(tuple(hashes))


def log(logger, msg, level=logging.DEBUG):
    if logger is not None:
        logger.log(level, msg)


def sort_by_central(set1, set2):
    """Determines a common index in two sets and returns a length 3
    tuple with the central index at the middle position and the two
    terminal indices as first and last indices."""
    central_set = set1 & set2
    union = set1 | set2
    assert len(central_set) == 1
    terminal1, terminal2 = union - central_set
    (central,) = central_set
    return (terminal1, central, terminal2), central


def merge_sets(fragments):
    """Merge a list of iterables."""
    # Hold the final fragments that can't be merged further, as they
    # contain distinct atoms.
    fragments = [frozenset(frag) for frag in fragments]
    merged = list()
    while len(fragments) > 0:
        popped = fragments.pop(0)
        # Look for an intersection between the popped unmerged fragment
        # and the remaining unmerged fragments.
        for frag in fragments:
            if popped & frag:
                fragments.remove(frag)
                # If a intersecting unmerged fragment is found merge
                # both fragments and append them at the end.
                fragments.append(popped | frag)
                break
        else:
            # Add the unmerged fragment into merged if it doesn't
            # intersect with any other unmerged fragment.
            merged.append(popped)
    return merged


def remove_duplicates(seq):
    tuples = [tuple(itm) for itm in seq]
    seen = set()
    seen_add = seen.add
    return [itm for itm in tuples if not (itm in seen or seen_add(itm))]


class OrderedEnum(Enum):
    def __ge__(self, other):
        if self.__class__ is other.__class__:
            return self.value >= other.value
        return NotImplemented

    def __gt__(self, other):
        if self.__class__ is other.__class__:
            return self.value > other.value
        return NotImplemented

    def __le__(self, other):
        if self.__class__ is other.__class__:
            return self.value <= other.value
        return NotImplemented

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented

    def __str__(self):
        return self.name


def timed(logger=None):
    def decorator(func):
        def wrapper(*args, **kwargs):
            start = time.time()
            result = func(*args, **kwargs)
            end = time.time()
            duration = end - start
            msg = f"Execution of '{func.__name__}' took {duration:.2f} s."
            if logger is not None:
                log(logger, msg)
            else:
                print(msg)
            return result

        return wrapper

    return decorator


def get_input(data, prompt, lbl_func=None):
    if lbl_func is None:
        lbl_func = lambda _: _
    labels = [lbl_func(d) for d in data]
    print(prompt)
    while True:
        for i, l in enumerate(labels):
            print(f"{i: >3d}: {l}")
        try:
            inp = int(input("Selection: "))
            if not (0 <= inp < len(labels)):
                raise ValueError
            break
        except ValueError:
            print("Invalid input!")
            print()
    return data[inp]


def expand(to_expand):
    if any([isinstance(to_expand, cls) for cls in (list, tuple, np.ndarray)]):
        return to_expand
    elif ".." in to_expand:
        start, end = [int(i) for i in to_expand.split("..")]
        return list(range(start, end))
    # Numbers
    else:
        return [int(to_expand)]


def full_expand(to_expand):
    try:
        split = to_expand.strip().split(",")
        expanded = list(it.chain(*[expand(te) for te in split]))
    except AttributeError:
        try:
            expanded = list(to_expand)
        except TypeError:
            expanded = [
                to_expand,
            ]
    return expanded


def file_or_str(*args, method=False, mode="r", exact=False):
    exts = args

    def inner_func(func):
        def wrapped(inp, *args, **kwargs):
            if method:
                obj = inp
                inp, *args = args
            p = Path(inp)
            looks_like_file = exts and (
                (p.suffix in exts) or (exact and p.name in exts)
            )
            if looks_like_file and p.is_file():
                with open(p, mode=mode) as handle:
                    inp = handle.read()
            elif looks_like_file and not p.exists():
                raise FileNotFoundError(
                    f"{inp} looks like a file/path, but it does not exist!"
                )
            if method:
                res = func(obj, inp, *args, **kwargs)
            else:
                res = func(inp, *args, **kwargs)
            return res

        return wrapped

    return inner_func


def recursive_update(d, u):
    """Recursive update of d with keys/values from u.

    From https://stackoverflow.com/questions/3232943
    """
    if u is None:
        return d

    # Best I can do is ... secretly transform hierarchical input
    # to old-style flat input.
    #
    # Try to transform new-style hierarchical input
    # to the old-style flat input. If this new-style proves useful
    # and everything is refactored accordingly the try/except could
    # be dropped. But now everything is still in the old-style.
    try:
        keys = list(u["type"].keys())
        if len(keys) == 1:
            key = keys[0]
            kwargs = u["type"][key]
            u["type"] = key
            try:
                u.update(kwargs)
            # Raised when kwargs is None, e.g., the input is hierarchical,
            # but no further keywords are provided for the given type.
            except TypeError:
                pass
    # Raised when u has no "type" key
    except KeyError:
        pass
    except AttributeError:
        pass

    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = recursive_update(d.get(k, {}), v)
        else:
            d[k] = v
    return d


def report_isotopes(geom, affect_str):
    if (geom.isotopes is not None) and len(geom.isotopes) > 0:
        print(f"Different isotopes were requested! This will affect {affect_str}.")
        atoms = geom.atoms
        masses = geom.masses
        for atom_ind, _ in geom.isotopes:
            print(f"\tAtom {atom_ind}{atoms[atom_ind]}: {masses[atom_ind]:.6f} au")
        print()


def report_frozen_atoms(geom):
    if hasattr(geom, "freeze_atoms") and len(geom.freeze_atoms) > 0:
        print(f"Coordinates of {len(geom.freeze_atoms)} atoms are frozen:")
        atoms = geom.atoms
        coords3d = geom.coords3d * BOHR2ANG
        fmt = " >12.8f"
        entries = list()
        for atom_ind in geom.freeze_atoms:
            atom = atoms[atom_ind]
            x, y, z = coords3d[atom_ind]
            entries.append(f"{atom_ind:03d} {atom} {x:{fmt}} {y:{fmt}} {z:{fmt}}")
        print("\t" + "  ".join(entries))
        print()


def highlight_text(text, width=80, level=0):
    levels = {
        #  horizontal
        #       vertical
        #           corner
        0: ("#", "#", "#"),
        1: ("-", "|", "+"),
    }
    full_length = len(text) + 4
    pad_len = width - full_length
    pad_len = (pad_len - (pad_len % 2)) // 2
    pad = " " * pad_len
    hchar, vchar, cornerchar = levels[level]
    full_row = cornerchar + (hchar * (full_length - 2)) + cornerchar
    highlight = (
        f"""{pad}{full_row}\n{pad}{vchar} {text.upper()} {vchar}\n{pad}{full_row}"""
    )
    return highlight


def interpolate_colors(values, c1, c2, num=32):
    """Expects two RGB colors c1 and c2."""
    c_diff = c2 - c1
    step = c_diff / (num - 1)
    colors = (c1 + np.arange(num)[:, None] * step).astype(int)

    # Map value interval onto interval range(num)
    # y = m*x + n
    val_min = values.min()
    val_max = values.max()
    m = abs((num - 1) / (val_min - val_max))
    n = -m * val_min
    inds = np.around(m * values + n).astype(int)
    rgb_colors = colors[inds]
    hex_colors = [f"#{r:02x}{g:02x}{b:02x}" for r, g, b in rgb_colors]
    return rgb_colors, hex_colors


def get_molecular_radius(coords3d, min_offset=0.9452):
    coords3d = coords3d.copy()
    mean = coords3d.mean(axis=0)
    coords3d -= mean[None, :]
    distances = np.linalg.norm(coords3d, axis=1)
    # When this function is utilized in a HardSphere calculator
    # the radii will be at least '2 * (2 * std)' apart,
    # as the std-offset will be added two times to all radii.
    std = max(min_offset, np.std(distances))
    radius = distances.mean() + 2 * std
    return radius


def filter_fixture_store(test_name):
    def inner_function(function):
        def wrapper(fixture_store):
            rb = fixture_store["results_bag"]
            filtered = {
                "results_bag": {k: v for k, v in rb.items() if k.startswith(test_name)}
            }
            return function(filtered)

        return wrapper

    return inner_function


def get_clock():
    ref = time.time()

    def clock(msg=""):
        nonlocal ref
        now = time.time()
        dur = now - ref
        ref = now
        print(f"{msg: >32}, {dur:.3f} s since last call!")

    return clock


def chunks(l, n):
    """Yield successive n-sized chunks from l.
    https://stackoverflow.com/a/312464
    """
    for i in range(0, len(l), n):
        yield l[i : i + n]


def describe(arr):
    shape = arr.shape
    min_ = arr.min()
    max_ = arr.max()
    mean = np.mean(arr)
    median = np.median(arr)
    var = np.var(arr)
    fmt = ".4f"
    return (
        f"shape={shape}, min={min_:{fmt}}, mean={mean:{fmt}}, "
        f"median={median:{fmt}}, max={max_:{fmt}}, variance={var:.4e}"
    )


def touch(fn):
    try:
        Path(fn).touch()
    except IsADirectoryError:
        pass


def approx_float(
    num: float, expected: float, abs_tol: float = 1e-6, rel_tol: float = 1e-12
) -> bool:
    def pos_or_none(num, name):
        assert (num > 0.0) or (num is None), f"{name} must be positive or None!"

    pos_or_none(abs_tol, "Absolute tolerance")
    pos_or_none(rel_tol, "Relative tolerance")
    assert abs_tol or rel_tol

    if rel_tol is None:
        rel_tol = 0.0
    rel_tol = rel_tol * expected
    tolerance = max(abs_tol, rel_tol)
    return abs(num - expected) <= tolerance


_CONV_FUNCS = {
    # First item in value converts to JSON dumpable type,
    # second items converts from JSON to original type.
    "energy": (lambda en: float(en), lambda en: float(en)),
    "all_energies": (
        lambda all_energies: all_energies.tolist(),
        lambda all_energies: np.array(all_energies, dtype=float),
    ),
    "forces": (
        lambda forces: forces.tolist(),
        lambda forces: np.array(forces, dtype=float).flatten(),
    ),
    "hessian": (
        lambda hessian: hessian.tolist(),
        lambda hessian: np.array(hessian, dtype=float),
    ),
}


def results_to_json(results):
    conv_results = {key: _CONV_FUNCS[key][0](val) for key, val in results.items()}
    return json.dumps(conv_results)


def json_to_results(as_json):
    results = {
        key: _CONV_FUNCS[key][1](val) for key, val in json.loads(as_json).items()
    }
    return results


def standard_state_corr(T=T_DEFAULT, p=p_DEFAULT, n=1):
    """dG for change of standard state from gas to solution of 1 mol/l"""
    Vm = n * R * T / p  # in m³
    Vm_litres = Vm * 1000
    dG = R * T * math.log(Vm_litres) / 1000  # kJ mol⁻¹
    dG_au = dG / AU2KJPERMOL
    return dG_au


def check_mem(mem, pal, avail_frac=0.85, logger=None):
    virt_mem = psutil.virtual_memory()
    mb_available = virt_mem.available * avail_frac / 1024 / 1024
    mb_requested = mem * pal
    msg = f"{mb_available:.2f} MB memory available, {mb_requested:.2f} MB requested."
    if mb_requested > mb_available:
        mb_corr = int(mb_available / pal)
        msg += f" Too much memory requested. Using smaller value of {mb_corr} MB."
    else:
        mb_corr = mem
    log(logger, msg)

    return mb_corr


def get_ratio(str_: str, comp_str: str) -> str:
    """See https://stackoverflow.com/a/17388505"""
    return SequenceMatcher(None, str_, comp_str).ratio()


def find_closest_sequence(str_: str, comp_strs: List[str]) -> Tuple[str, float]:
    ratios = [get_ratio(str_, comp_str) for comp_str in comp_strs]
    argmax = np.argmax(ratios)
    max_ratio = ratios[argmax]
    best_match = comp_strs[argmax]
    return (best_match, max_ratio)


def increment_fn(org_fn: str, suffix: Optional[str] = None) -> str:
    """
    Append, or increase a suffixed counter on a given filename.
    If no counter is present it will be set to zero. Otherwise
    it is incremented by one.

    >>> increment_fn("opt", "rebuilt")
    'opt_rebuilt_000'
    >>> increment_fn("opt")
    'opt_000'
    >>> increment_fn("opt_rebuilt_000", "rebuilt")
    'opt_rebuilt_001'

    Parameters
    ----------
    org_fn
        The original, unaltered filename.
    suffix
        Optional suffix to be append.

    Returns
    -------
    incr_fn
        Modified filename with optional suffix and incremented counter.
    """
    if suffix is not None and len(suffix) > 0:
        suffix = f"_{suffix}"
    else:
        suffix = ""

    # Check if prefix is present
    regex = re.compile(rf"(.+)({suffix}_(\d+))")
    if mobj := regex.match(org_fn):
        org_fn, _, counter = mobj.groups()
        counter = int(counter)
        counter += 1
    else:
        counter = 0

    incr_fn = f"{org_fn}{suffix}_{counter:03d}"
    return incr_fn


def get_box(coords3d, offset=0.0):
    mins = coords3d.min(axis=0)
    maxs = coords3d.max(axis=0)
    box = np.stack((mins, maxs), axis=1)
    box[:, 0] -= offset
    box[:, 1] += offset
    return box


def molecular_volume(
    coords3d: NDArray[float],
    vdw_radii: NDArray[float],
    n_trial: int = 10_000,
    offset: float = 1.0,
) -> Tuple[float, float, float]:
    """Monte-Carlo estimate of molecular volume using Van der Waals spheres.
    Cartesian coordinates and VdW-radii are expected in Bohr!
    """
    box = get_box(coords3d, offset=offset)
    edges = np.diff(box, axis=1).flatten()
    box_volume = abs(np.prod(edges))
    # print(f"Box with volume {box_volume:.2f} a0³ and dimensions\n{box}")

    def trial_points(n):
        tps = np.random.rand(n, 3)
        box_min = box[:, 0]
        return box_min + tps * edges

    tps = trial_points(n=n_trial)

    dists = np.linalg.norm(coords3d[:, None, :] - tps, axis=2)
    below_radius = dists <= vdw_radii[:, None]
    below_radius = below_radius.sum()
    ratio = below_radius / n_trial
    mol_vol = ratio * box_volume  # a0³ / Molecule
    mol_vol_ang3 = mol_vol * BOHR2ANG**3  # Å³ / Molecule
    molar_vol = mol_vol_ang3 * NA * 1e-24  # l/mol
    return mol_vol, mol_vol_ang3, molar_vol


def get_cubic_crystal(l, n=2, atom="Na"):
    coords3d = list()
    atoms = list()
    for xn in range(-n, n + 1):
        for yn in range(-n, n + 1):
            for zn in range(-n, n + 1):
                coords3d.append((l * xn, l * yn, l * zn))
                atoms.append(atom)
    coords3d = np.array(coords3d)
    return atoms, coords3d


# Unicode subscript characters for numbers
_NUM_SUBSCRIPTS = {str(i): chr(int(f"0x208{i}", 16)) for i in range(10)}


def to_subscript_num(num):
    return "".join([_NUM_SUBSCRIPTS[c] for c in str(num)])


def to_sets(iterable):
    return set([frozenset(i) for i in iterable])


def cache_arrays(sources, dest):
    dest_path = Path(dest).with_suffix(".npz")

    def cache_decorator(func):
        def wrapped_func(*args, **kwargs):
            if dest_path.exists():
                npzfile = np.load(dest_path)
                arrays = tuple([npzfile[src] for src in sources])
            else:
                arrays = func(*args, **kwargs)
                assert len(arrays) == len(sources)
                kwds = {src: arr for src, arr in zip(sources, arrays)}
                np.savez(dest_path, **kwds)
            return arrays

        return wrapped_func

    return cache_decorator


def estimate(gen, elems):
    tot_dur = 0.0
    for i in range(1, elems + 1):
        start = time.time()
        elem = next(gen)
        dur = time.time() - start
        tot_dur += dur
        ran_ratio = i / elems
        elems_left = elems - i
        dur_per_elem = tot_dur / i
        est_dur = dur_per_elem * elems_left
        est_dur_min = est_dur / 60
        print(f"{ran_ratio: >8.2%} finished ... {est_dur_min: >8.2f} min left")
        yield elem


def get_random_path(stem=""):
    if stem != "":
        stem += "_"
    while (path := Path(f"{stem}{uuid.uuid1()}")).exists():
        pass
    return path


def kill_dir(path):
    """Innocent function remove a directory.

    It must contain only files and no other directories.
    So this won't do too much damage hopefully.
    """
    for fn in Path(path).iterdir():
        fn.unlink()
    path.rmdir()


def rms(arr):
    if isinstance(arr, torch.Tensor):
        result = torch.sqrt(torch.mean(arr**2))
    else:
        result = np.sqrt(np.mean(arr**2))
    return result
